from celery.utils.log import get_logger
import sleek_lvalert

log = get_logger(__name__)


class LVAlertClient(sleek_lvalert.LVAlertClient):

    def __init__(self, *args, nodes=None, **kwargs):
        super().__init__(*args, **kwargs)
        self._needed_subscriptions = set(nodes or ())
        self.add_event_handler('session_start', self._resubscribe)

    def _resubscribe(self, event):
        log.info('Resubscribing to PubSub nodes')
        current_subscriptions = set(self.get_subscriptions())
        self.subscribe(*(self._needed_subscriptions - current_subscriptions))



import configparser
import os
import subprocess

import pytest
from requests.exceptions import HTTPError
from unittest.mock import Mock

from .. import app
from ..tasks import condor
from ..tasks import inference


def mock_find_urls(available_frametypes):
    """Return mock gwdatafind.find_urls, which returns 'path_to_data' when
    input ifo and frametype are in available_frametypes.

    available_frametypes : dictionary
        Dictionary whose key is ifo and item is frametype
    """
    _available_frametypes = dict(
        (ifo[0], frametype) for ifo, frametype in available_frametypes.items())

    def _mock_find_url(ifo, frametype, start, end):
        data_exists = (ifo in _available_frametypes.keys() and
                       frametype == _available_frametypes[ifo])
        if data_exists:
            return ['path_to_data']
        else:
            return []

    return _mock_find_url


@pytest.mark.parametrize(
    'available_frametypes,answer',
    [[{}, None],
     [{'H1': app.conf['low_latency_frame_types']['H1']}, None],
     [app.conf['low_latency_frame_types'],
      app.conf['low_latency_frame_types']],
     [app.conf['high_latency_frame_types'],
      app.conf['high_latency_frame_types']]]
)
def test_query_data(monkeypatch, available_frametypes, answer):
    monkeypatch.setattr('gwcelery.tasks.inference.find_urls',
                        mock_find_urls(available_frametypes))
    if answer is None:
        with pytest.raises(inference.NotEnoughData):
            inference.query_data(1187008882)
    else:
        assert inference.query_data(1187008882) == answer


def test_upload_no_frametypes(monkeypatch):
    upload = Mock()
    monkeypatch.setattr('gwcelery.tasks.gracedb.upload.run', upload)
    inference.upload_no_frame_files(
        None, inference.NotEnoughData(), 'test', 'S1234')
    upload.assert_called_once()


def test_find_appropriate_cal_env(tmp_path):
    filepath = tmp_path / 'H_CalEnvs.txt'
    t1, cal1 = 1126259462, 'O1_calibration.txt'
    t2, cal2 = 1187008882, 'O2_calibration.txt'
    filepath.write_text('{t1} {cal1}\n{t2} {cal2}'.format(
            t1=t1, cal1=cal1, t2=t2, cal2=cal2))
    dir_name = str(tmp_path)
    trigtime_answer = [
        (t1 - 1, cal1), ((t1 + t2) / 2., cal1), (t2 + 1, cal2)]
    for trigtime, answer in trigtime_answer:
        assert os.path.basename(
            inference._find_appropriate_cal_env(trigtime, dir_name)) == answer


@pytest.mark.parametrize(
    'mc,q,sid,answers',
    [[1., 1., None,
      [('analysis', 'engine', 'lalinferencenest'),
       ('analysis', 'nparallel', '4'),
       ('analysis', 'roq', 'True'),
       ('paths', 'roq_b_matrix_directory',
        '/home/cbc/ROQ_data/IMRPhenomPv2/'),
       ('engine', 'approx', 'IMRPhenomPv2pseudoFourPN'),
       ('bayeswave', 'Nchain', '10'),
       ('bayeswave', 'Niter', '100000'),
       ('condor', 'bayeswave_request_memory', '16000'),
       ('condor', 'bayeswavepost_request_memory', '16000'),
       ('bayeswave', 'bw_srate', '4096')]],
     [10., 1., None,
      [('condor', 'bayeswave_request_memory', '1000'),
       ('condor', 'bayeswavepost_request_memory', '4000')]],
     [10., 1., 'S1234', []],
     [1., 1. / 10., None,
      [('analysis', 'engine', 'lalinferencenest'),
       ('analysis', 'nparallel', '4'),
       ('analysis', 'roq', 'True'),
       ('paths', 'roq_b_matrix_directory',
        '/home/cbc/ROQ_data/SEOBNRv4ROQ/'),
       ('engine', 'approx', 'SEOBNRv4_ROMpseudoFourPN')]],
     [1., 1. / 20., None,
      [('analysis', 'engine', 'lalinferencemcmc'),
       ('analysis', 'nparallel', '10'),
       ('analysis', 'roq', 'False'),
       ('engine', 'approx', 'SEOBNRv4_ROMpseudoFourPN'),
       ('engine', 'neff', '300')]]]
)
def test_prepare_ini(monkeypatch, mc, q, sid, answers):
    upload = Mock()

    def mock_calenv(trigtime, path):
        return 'path_to_calenv'

    monkeypatch.setattr('gwcelery.tasks.gracedb.upload.run', upload)
    monkeypatch.setattr(
        'gwcelery.tasks.inference._find_appropriate_cal_env',
        mock_calenv)
    monkeypatch.setitem(app.conf, 'gracedb_host', 'gracedb.ligo.org')

    m1 = mc * q**(-3. / 5.) * (1 + q)**(1. / 5.)
    m2 = q * m1
    event = {
        'gpstime': 1187008882,
        'graceid': 'G1234',
        'extra_attributes':
            {'SingleInspiral': [{'mass1': m1, 'mass2': m2, 'mchirp': mc}]}
    }
    config = configparser.ConfigParser()
    config.read_string(
        inference.prepare_ini(app.conf['low_latency_frame_types'], event, sid))
    for section, key, answer in answers:
        assert config[section][key] == answer

    if sid is not None:
        upload.assert_called_once()


def test_pre_pe_tasks(monkeypatch):
    event, sid = {'gpstime': 1187008882}, 'S1234'
    frametype_dict = app.conf['low_latency_frame_types']

    def mock_query_data(gpstime):
        return frametype_dict

    def mock_prepare_ini(f, e, s):
        assert f == frametype_dict
        assert e == event
        assert s == sid

    query_data = Mock(side_effect=mock_query_data)
    prepare_ini = Mock(side_effect=mock_prepare_ini)
    monkeypatch.setattr('gwcelery.tasks.inference.query_data.run', query_data)
    monkeypatch.setattr('gwcelery.tasks.inference.prepare_ini.run',
                        prepare_ini)

    inference.pre_pe_tasks(event, sid).delay()
    query_data.assert_called_once()
    prepare_ini.assert_called_once()


@pytest.mark.parametrize('psd', [b'psd', None])
def test_setup_dag_for_lalinference_(monkeypatch, tmp_path, psd):
    ini, coinc = 'ini', b'coinc'
    rundir = str(tmp_path)
    dag = 'lalinference dag'

    def _subprocess_run(cmd, **kwargs):
        assert cmd[0] == 'lalinference_pipe'
        del cmd[0]
        assert '--run-path' in cmd and '--coinc' in cmd
        idx = cmd.index('--run-path')
        assert cmd[idx + 1] == rundir
        del cmd[idx:idx + 2]
        idx = cmd.index('--coinc')
        path_to_coinc = cmd[idx + 1]
        with open(path_to_coinc, 'rb') as f:
            assert f.read() == coinc
        del cmd[idx:idx + 2]
        if psd is not None:
            assert '--psd' in cmd
            idx = cmd.index('--psd')
            path_to_psd = cmd[idx + 1]
            with open(path_to_psd, 'rb') as f:
                assert f.read() == psd
            del cmd[idx:idx + 2]
        path_to_ini, = cmd
        with open(os.path.join(rundir, 'multidag.dag'), 'w') as f:
            f.write(dag)

    monkeypatch.setattr('subprocess.run', _subprocess_run)
    path_to_dag = inference._setup_dag_for_lalinference(
        (coinc, psd), ini, rundir, 'S1234')
    with open(path_to_dag, 'r') as f:
        assert f.read() == dag


@pytest.mark.parametrize(
    'host', ['gracedb-playground.ligo.org', 'gracedb.ligo.org'])
def test_setup_dag_for_bilby(monkeypatch, tmp_path, host):
    event, rundir = {}, str(tmp_path)
    pid, sid = 'G1234', 'S1234'
    monkeypatch.setitem(app.conf, 'gracedb_host', host)
    dag = 'bilby dag'

    def _subprocess_run(cmd, **kwargs):
        is_quick_conf = 'o2replay' in cmd and 'FastTest' in cmd
        if host == 'gracedb.ligo.org':
            assert not is_quick_conf
        else:
            assert is_quick_conf
        with open(os.path.join(rundir, 'bilby_config.ini'), 'w') as f:
            f.write('bilby configuration ini file')
        dir = os.path.join(rundir, 'submit')
        os.mkdir(dir)
        with open(os.path.join(dir, 'dag_bilby.submit'), 'w') as f:
            f.write(dag)

    upload = Mock()
    monkeypatch.setattr('subprocess.run', _subprocess_run)
    monkeypatch.setattr('gwcelery.tasks.gracedb.upload.run', upload)

    path_to_dag = inference._setup_dag_for_bilby(event, rundir, pid, sid)
    with open(path_to_dag, 'r') as f:
        assert f.read() == dag
    upload.assert_called_once()


@pytest.mark.parametrize('pipeline', ['lalinference', 'bilby'])
def test_setup_dag_for_failure(monkeypatch, tmp_path, pipeline):
    upload = Mock()
    monkeypatch.setattr('gwcelery.tasks.gracedb.upload.run', upload)
    monkeypatch.setattr('subprocess.run',
                        Mock(side_effect=subprocess.CalledProcessError(
                            1, ['cmd'], output=b'out', stderr=b'err')))

    rundir = str(tmp_path)

    with pytest.raises(subprocess.CalledProcessError):
        if pipeline == 'lalinference':
            inference._setup_dag_for_lalinference(
                (b'coinc', b'psd'), 'ini', rundir, 'S1234')
        elif pipeline == 'bilby':
            inference._setup_dag_for_bilby({}, rundir, 'G1234', 'S1234')
    assert not os.path.exists(rundir)
    upload.assert_called_once()


@pytest.mark.parametrize('pipeline',
                         ['lalinference', 'bilby', 'my_awesome_pipeline'])
def test_dag_prepare_task(monkeypatch, pipeline):
    gid, sid = 'G1234', 'S1234'
    coinc, psd, ini, event = b'coinc', b'psd', 'ini', {'gpstime': 1187008882}
    rundir = 'rundir'
    path_to_dag = os.path.join(rundir, 'parameter_estimation.dag')

    def mock_download(filename, gid):
        if filename == 'coinc.xml':
            return coinc
        elif filename == 'psd.xml.gz':
            return psd

    def _setup_dag_for_lalinference(c_p, i, r, s):
        c, p = c_p
        assert (c == coinc and p == psd and i == ini
                and r == rundir and s == sid)
        return path_to_dag

    def _setup_dag_for_bilby(e, r, p, s):
        assert e == event and r == rundir and p == gid and s == sid
        return path_to_dag

    def _subprocess_run(cmd, **kwargs):
        assert cmd == ['condor_submit_dag', '-no_submit', path_to_dag]

    mock_setup_dag_for_lalinference = \
        Mock(side_effect=_setup_dag_for_lalinference)
    mock_setup_dag_for_bilby = Mock(side_effect=_setup_dag_for_bilby)
    monkeypatch.setattr('gwcelery.tasks.gracedb.download.run', mock_download)
    monkeypatch.setattr('gwcelery.tasks.gracedb.get_event.run',
                        Mock(return_value=event))
    monkeypatch.setattr(
        'gwcelery.tasks.inference._setup_dag_for_lalinference.run',
        mock_setup_dag_for_lalinference)
    monkeypatch.setattr('gwcelery.tasks.inference._setup_dag_for_bilby.run',
                        mock_setup_dag_for_bilby)
    monkeypatch.setattr('subprocess.run', _subprocess_run)
    if pipeline in ['lalinference', 'bilby']:
        inference.dag_prepare_task(rundir, sid, gid, pipeline, ini).delay()
        if pipeline == 'lalinference':
            mock_setup_dag_for_lalinference.assert_called_once()
        elif pipeline == 'bilby':
            mock_setup_dag_for_bilby.assert_called_once()
    else:
        with pytest.raises(NotImplementedError):
            inference.dag_prepare_task(rundir, sid, gid, pipeline, ini).delay()


@pytest.mark.parametrize('exc', [condor.JobAborted(1, 'test'),
                                 condor.JobFailed(1, 'test')])
def test_job_error_notification(monkeypatch, tmp_path, exc):
    filenames = ['pe.log', 'pe.err', 'pe.out']
    for filename in filenames:
        with open(str(tmp_path / filename), 'w') as f:
            f.write('test')
    upload = Mock()
    monkeypatch.setattr('gwcelery.tasks.gracedb.upload.run', upload)

    inference.job_error_notification(
        None, exc, 'test', 'S1234', str(tmp_path), 'lalinference')
    assert upload.call_count == len(filenames) + 1


@pytest.mark.parametrize('pipeline',
                         ['lalinference', 'bilby', 'my_awesome_pipeline'])
def test_upload_url(monkeypatch, tmp_path, pipeline):
    upload = Mock()
    monkeypatch.setattr('gwcelery.tasks.gracedb.upload.run', upload)
    if pipeline in ['lalinference', 'bilby']:
        if pipeline == 'lalinference':
            path = str(tmp_path / 'posplots.html')
        else:
            path = str(tmp_path / 'home.html')
        with open(path, 'w') as f:
            f.write('test')
        inference._upload_url(str(tmp_path), 'G1234', pipeline)
        upload.assert_called_once()
    else:
        with pytest.raises(NotImplementedError):
            inference._upload_url(str(tmp_path), 'G1234', pipeline)


@pytest.mark.parametrize('pipeline',
                         ['lalinference', 'bilby', 'my_awesome_pipeline'])
def test_dag_finished(monkeypatch, tmp_path, pipeline):
    gid = 'G1234'
    rundir = str(tmp_path / 'rundir')
    resultdir = str(tmp_path / 'rundir/result')
    pe_results_path = str(tmp_path / 'public_html/online_pe')
    monkeypatch.setitem(app.conf, 'pe_results_path', pe_results_path)
    pe_results_path = os.path.join(pe_results_path, gid, pipeline)
    os.makedirs(rundir)
    os.makedirs(resultdir)
    os.makedirs(pe_results_path)

    upload = Mock()
    _upload_url = Mock()
    create_label = Mock()
    monkeypatch.setattr('gwcelery.tasks.gracedb.upload.run', upload)
    monkeypatch.setattr('gwcelery.tasks.inference._upload_url.run',
                        _upload_url)
    monkeypatch.setattr('gwcelery.tasks.gracedb.create_label.run',
                        create_label)

    if pipeline in ['lalinference', 'bilby']:
        if pipeline == 'lalinference':
            paths = [os.path.join(rundir,
                                  'lalinference_1187008756-1187008882.dag'),
                     os.path.join(rundir, 'glitch_median_PSD_forLI_H1.dat'),
                     os.path.join(rundir, 'glitch_median_PSD_forLI_L1.dat'),
                     os.path.join(rundir, 'glitch_median_PSD_forLI_V1.dat'),
                     os.path.join(rundir, 'posterior_samples.hdf5'),
                     os.path.join(pe_results_path, 'extrinsic.png'),
                     os.path.join(pe_results_path, 'sourceFrame.png')]
        else:
            paths = [os.path.join(resultdir, 'bilby_merge_result.json'),
                     os.path.join(resultdir, 'bilby_extrinsic_corner.png'),
                     os.path.join(resultdir, 'bilby_intrinsic_corner.png')]
        for path in paths:
            with open(path, 'wb') as f:
                f.write(b'result')

        inference.dag_finished(rundir, gid, 'S1234', pipeline)
        assert upload.call_count == len(paths)
        _upload_url.assert_called_once()
        if pipeline == 'lalinference':
            create_label.assert_called_once()
        else:
            create_label.assert_not_called()
        assert not os.path.exists(rundir)
    else:
        with pytest.raises(NotImplementedError):
            inference.dag_finished(rundir, gid, 'S1234', pipeline)


def test_download_psd_failure(monkeypatch):
    monkeypatch.setattr('gwcelery.tasks.gracedb.download',
                        Mock(side_effect=HTTPError))
    assert inference._download_psd('G1234') is None


def test_start_pe(monkeypatch, tmp_path):
    path_to_sub = 'pe.dag.condor.sub'

    @app.task
    def mock_task():
        return path_to_sub

    dag_prepare_task = Mock(return_value=mock_task.s())

    def mock_condor_submit(path):
        assert path == path_to_sub

    condor_submit = Mock(side_effect=mock_condor_submit)
    dag_finished = Mock()
    monkeypatch.setattr('gwcelery.tasks.gracedb.upload.run', Mock())
    monkeypatch.setattr('distutils.dir_util.mkpath',
                        Mock(return_value=str(tmp_path)))
    monkeypatch.setattr('gwcelery.tasks.inference.dag_prepare_task',
                        dag_prepare_task)
    monkeypatch.setattr('gwcelery.tasks.condor.submit.run', condor_submit)
    monkeypatch.setattr('gwcelery.tasks.inference.dag_finished.run',
                        dag_finished)

    inference.start_pe('ini', 'G1234', 'S1234', 'lalinference')
    dag_prepare_task.assert_called_once()
    condor_submit.assert_called_once()
    dag_finished.assert_called_once()



"""Create mock events from the "First Two Years" paper."""
import io
import random

from celery.task import PeriodicTask
from celery.utils.log import get_task_logger
from glue.ligolw import utils
from glue.ligolw import lsctables
import lal
from ligo.skymap.io.events.ligolw import ContentHandler
import numpy as np
import pkg_resources

from ..import app
from . import gracedb

log = get_task_logger(__name__)


def pick_coinc():
    """Pick a coincidence from the "First Two Years" paper."""
    filename = pkg_resources.resource_filename(
        __name__, '../data/first2years/2016/gstlal.xml.gz')
    xmldoc = utils.load_filename(filename, contenthandler=ContentHandler)
    root, = xmldoc.childNodes

    # Remove unneeded tables
    for lsctable in (
            lsctables.FilterTable,
            lsctables.SegmentTable,
            lsctables.SegmentDefTable,
            lsctables.SimInspiralTable,
            lsctables.SummValueTable,
            lsctables.SearchSummVarsTable):
        root.removeChild(lsctable.get_table(xmldoc))

    coinc_inspiral_table = table = lsctables.CoincInspiralTable.get_table(
        xmldoc)

    # Determine event with most recent sideral time
    gps_time_now = lal.GPSTimeNow()
    gmsts = np.asarray([lal.GreenwichMeanSiderealTime(_.end) for _ in table])
    gmst_now = lal.GreenwichMeanSiderealTime(gps_time_now)
    div, rem = divmod(gmst_now - gmsts, 2 * np.pi)
    i = np.argmin(rem)
    new_gmst = div[i] * 2 * np.pi + gmsts[i]
    old_time = table[i].end
    new_time = lal.LIGOTimeGPS()
    result = lal.GreenwichMeanSiderealTimeToGPS(new_gmst, new_time)
    result.disown()
    del result
    delta_t = new_time - old_time
    target_coinc_event_id = int(table[i].coinc_event_id)

    # Remove unneeded rows
    table[:] = [row for row in table
                if int(row.coinc_event_id) == target_coinc_event_id]
    target_end_time = table[0].get_end()

    coinc_table = table = lsctables.CoincTable.get_table(xmldoc)
    table[:] = [row for row in table
                if int(row.coinc_event_id) == target_coinc_event_id]

    table = lsctables.CoincMapTable.get_table(xmldoc)
    table[:] = [row for row in table
                if int(row.coinc_event_id) == target_coinc_event_id]
    target_sngl_inspirals = frozenset(row.event_id for row in table)

    sngl_inspiral_table = table = lsctables.SnglInspiralTable.get_table(xmldoc)
    table[:] = [row for row in table if row.event_id in target_sngl_inspirals]

    table = lsctables.ProcessTable.get_table(xmldoc)
    table[:] = [row for row in table if row.program == 'gstlal_inspiral']
    target_process_ids = frozenset(row.process_id for row in table)

    table = lsctables.SearchSummaryTable.get_table(xmldoc)
    table[:] = [row for row in table if target_end_time in row.get_out()
                and row.process_id in target_process_ids]
    target_process_ids = frozenset(row.process_id for row in table)

    table = lsctables.ProcessTable.get_table(xmldoc)
    table[:] = [row for row in table if row.process_id in target_process_ids]

    table = lsctables.ProcessParamsTable.get_table(xmldoc)
    table[:] = [row for row in table if row.process_id in target_process_ids]

    # Shift event times
    for row in coinc_inspiral_table:
        row.end += delta_t
    for row in sngl_inspiral_table:
        row.end += delta_t
        row.end_time_gmst = lal.GreenwichMeanSiderealTime(row.end)

    # The old version of gstlal used to produce the "First Two Years" data set
    # stored likelihood in the coinc_event.likelihood column, but newer
    # versions store the *natural log* of the likelihood here. The p_astro
    # calculation requires this to be log likelihood.
    for row in coinc_table:
        row.likelihood = np.log(row.likelihood)

    # Gstlal stores the template's SVD bank index in the Gamma1 column.
    # Fill this in so that we can calculate p_astro
    # (see :mod:`gwcelery.tasks.p_astro_gstlal`).
    for row in sngl_inspiral_table:
        row.Gamma1 = 16

    coinc_xml = io.BytesIO()
    utils.write_fileobj(xmldoc, coinc_xml)
    return coinc_xml.getvalue()


def _jitter_snr(coinc_bytes):
    coinc_xml = io.BytesIO(coinc_bytes)
    xmldoc, _ = utils.load_fileobj(coinc_xml, contenthandler=ContentHandler)

    coinc_inspiral_table = lsctables.CoincInspiralTable.get_table(xmldoc)

    # Add a tiny amount of jitter in SNR so that uploads have random
    # preferred event precedence.
    for row in coinc_inspiral_table:
        row.snr += random.gauss(0, 1e-9)

    coinc_xml = io.BytesIO()
    utils.write_fileobj(xmldoc, coinc_xml)
    return coinc_xml.getvalue()


@app.task(shared=False)
def _vet_event(superevents):
    if superevents:
        gracedb.create_signoff.s(
            random.choice(['NO', 'OK']),
            'If this had been a real gravitational-wave event candidate, '
            'then an on-duty scientist would have left a comment here on '
            'data quality and the status of the detectors.',
            'ADV', superevents[0]['superevent_id']
        ).apply_async()


@gracedb.task(ignore_result=True, shared=False)
def _upload_psd(graceid):
    psd = pkg_resources.resource_string(
        __name__, '../data/first2years/2016/psd.xml.gz')
    gracedb.upload(psd, 'psd.xml.gz', graceid, 'Noise PSD', ['psd'])


@app.task(base=PeriodicTask, shared=False, run_every=3600)
def upload_event():
    """Upload a random event from the "First Two Years" paper.

    After 2 minutes, randomly either retract or confirm the event to send a
    retraction or initial notice respectively.
    """
    coinc = pick_coinc()

    graceid = gracedb.create_event(coinc, 'MDC', 'gstlal', 'CBC')
    log.info('uploaded as %s', graceid)

    if app.conf['mock_events_simulate_multiple_uploads']:
        num = 15
        for _ in range(num):
            (
                gracedb.create_event.s(
                    _jitter_snr(coinc), 'MDC', 'gstlal', 'CBC'
                )
                |
                _upload_psd.s()
            ).apply_async()

    (
        _upload_psd.si(graceid)
        |
        gracedb.get_superevents.si(
            'MDC event: {}'.format(graceid)
        ).set(countdown=600)
        |
        _vet_event.s()
    ).apply_async()



import os
import sys
from unittest.mock import Mock

from ..tools.condor_submit_helper import main


def test_condor_submit_helper(monkeypatch, tmpdir):

    # Create a fake machine ClassAd filename.
    filename = str(tmpdir / 'classad')
    with open(filename, 'w') as f:
        # "This mission is too important for me to allow you to jeopardize it."
        print('ClientMachine = "hal9000.discovery"', file=f)
        # "Open the pod bay doors, HAL."
        print('RemoteUser = "david.bowman"', file=f)
        # "I know that you and Frank were planning to disconnect me,
        # and I'm afraid that's something I cannot allow to happen."
        print('Activity = killing', file=f)

    monkeypatch.setenv('_CONDOR_MACHINE_AD', filename)
    monkeypatch.setenv('CELERY_BROKER_URL', '')
    mock_execvp = Mock()
    monkeypatch.setattr(sys, 'argv', ['condor_submit_helper', 'foo', 'bar'])
    monkeypatch.setattr(os, 'execvp', mock_execvp)

    main()

    assert os.environ['CELERY_BROKER_URL'] == 'redis://hal9000.discovery'
    mock_execvp.assert_called_once_with('foo', ['foo', 'bar'])



from celery import exceptions
import pytest
from unittest.mock import call, patch, Mock

from requests.exceptions import HTTPError
from requests.models import Response

from ..tasks import gracedb, superevents
from ..util import resource_json

lvalert_content = {
    'object': {
        'graceid': '',
        'gpstime': '',
        'pipeline': '',
        'group': '',
        'created': '',
        'far': '',
        'instruments': '',
        'labels': [],
        'extra_attributes': [],
        'submitter': 'deep.chatterjee@ligo.org',
        'offline': False
    },
    'data': {
        'graceid': '',
        'gpstime': '',
        'pipeline': '',
        'group': '',
        'created': '',
        'far': '',
        'instruments': '',
        'labels': [],
        'extra_attributes': [],
        'submitter': 'deep.chatterjee@ligo.org',
        'offline': False
    },
    'alert_type': 'new',
    'uid': ''
}


G330308_RESPONSE = {
    'graceid': 'G330308',
    'gpstime': 1239917954.250977,
    'pipeline': 'pycbc',
    'group': 'CBC',
    'offline': False,
    'far': 1.48874654585461e-08,
    'instruments': 'H1,L1',
    'extra_attributes': {
        'SingleInspiral': [
            {
                'chisq': 1.07,
                'snr': 7.95,
                'ifo': 'L1'
            },
            {
                'chisq': 0.54,
                'snr': 6.35,
                'ifo': 'H1'
            }
        ],
        'CoincInspiral': {
            'snr': 10.17
        }
    },
    'search': 'AllSky',
    'superevent': 'S190421ar',
    'labels': ['PASTRO_READY', 'EMBRIGHT_READY']
}


G330298_RESPONSE = {
    'graceid': 'G330298',
    'gpstime': 1239917954.40918,
    'pipeline': 'spiir',
    'group': 'CBC',
    'offline': False,
    'far': 5.57979637960671e-06,
    'instruments': 'H1,L1',
    'extra_attributes': {
        'SingleInspiral': [
            {
                'chisq': 0.61,
                'snr': 6.64,
                'ifo': 'L1'
            },
            {
                'chisq': 1.1,
                'snr': 8.14,
                'ifo': 'H1'
            }
        ],
        'CoincInspiral': {
            'snr': 10.51
        }
    },
    'search': 'HighMass',
    'superevent': 'S190421ar',
    'labels': []
}


G000003_RESPONSE = {
    "graceid": "G000003",
    "pipeline": "gstlal",
    "group": "CBC",
    "search": "AllSky",
    "far": 1.e-6,
    "superevent": "S100",
    "group": "CBC",
    "instruments": "H1,L1",
    'extra_attributes': {
        'SingleInspiral': [
            {
                'chisq': 1.07,
                'snr': 10.95,
                'ifo': 'L1'
            },
            {
                'chisq': 0.54,
                'snr': 12.35,
                'ifo': 'H1'
            }
        ],
        'CoincInspiral': {
            'snr': 10.17
        },
    },
    "offline": False,
    "gpstime": 1000000,
    "labels": ["RAVEN_ALERT", "PASTRO_READY",
               "EMBRIGHT_READY", "SKYMAP_READY"]
}


def assert_not_called_with(mock, *args, **kwargs):
    assert call(*args, **kwargs) not in mock.call_args_list


def s100response(sid):
    return {
        "category": "Production",
        "created": "2018-09-19 18:28:54 UTC",
        "em_events": ["E0155"],
        "far": 0.0000342353,
        "gw_events": [
            "G000003"
        ],
        "gw_id": None,
        "labels": [],
        "preferred_event": "G000003",
        "submitter": "albert.einstein@LIGO.ORG",
        "superevent_id": "S100",
        "t_0": 1167264018.0,
        "t_end": 1167264025.0,
        "t_start": 1167264014.0
    }


@pytest.fixture(autouse=True)
def mock_db(monkeypatch):

    def get_superevents(*args, **kwargs):
        return resource_json(__name__, 'data/superevents.json')['superevents']

    def get_event(gid):
        if gid == "T0212":
            return resource_json(__name__, 'data/T0212_S0039_preferred.json')
        elif gid == "G000003":
            return G000003_RESPONSE
        elif gid == "G000012":
            return resource_json(__name__, 'data/G000012_S0040_preferred.json')
        elif gid == "G330308":
            return G330308_RESPONSE
        elif gid == "G330298":
            return G330298_RESPONSE
        else:
            raise ValueError("Called with incorrect preferred event %s" % gid)

    monkeypatch.setattr('gwcelery.tasks.gracedb.create_superevent', Mock())
    monkeypatch.setattr('gwcelery.tasks.gracedb.get_superevents', Mock())
    monkeypatch.setattr('gwcelery.tasks.gracedb.add_event_to_superevent',
                        Mock())
    monkeypatch.setattr('gwcelery.tasks.gracedb.get_superevents',
                        get_superevents)
    monkeypatch.setattr('gwcelery.tasks.gracedb.get_event', get_event)


def test_select_preferred_event():
    """Provide the list of trigger information of a sample event."""
    events = resource_json(__name__, 'data/sample_events.json')
    r = superevents.select_preferred_event(events)
    assert r['graceid'] == 'G3'


@pytest.mark.parametrize(
    'superevent_labels,new_event_labels,preferred_event_labels,new_event_id',
    [[[],
      ['EMBRIGHT_READY', 'PASTRO_READY', 'SKYMAP_READY'],
      ['EMBRIGHT_READY', 'PASTRO_READY'], 'T1234'],
     [[],
      ['EMBRIGHT_READY', 'PASTRO_READY'],
      ['EMBRIGHT_READY', 'PASTRO_READY', 'SKYMAP_READY'], 'T1234'],
     [[],
      ['EMBRIGHT_READY', 'PASTRO_READY',
       'SKYMAP_READY'],
      ['EMBRIGHT_READY', 'PASTRO_READY',
       'SKYMAP_READY'], 'T1234'],
     [[],
      ['EMBRIGHT_READY', 'PASTRO_READY'],
      ['EMBRIGHT_READY', 'PASTRO_READY',
       'SKYMAP_READY'], 'T0212'],
     [['EM_Selected', 'ADVREQ', 'DQOK'],
      ['EMBRIGHT_READY', 'PASTRO_READY',
       'SKYMAP_READY'],
      ['EMBRIGHT_READY', 'PASTRO_READY',
       'SKYMAP_READY'], 'T1234']])
def test_update_preferred_event(superevent_labels, new_event_labels,
                                preferred_event_labels, new_event_id):
    """Test scenarios pertaining to superevent S0039, with a
    preferred event T0212. The new event dictionary corresponds
    to either T0212, when new labels are added, or a new event
    T1234. The new event T1234 passes FAR threshold and has higher
    SNR compared to the preferred event.
    """
    preferred_event_dictionary = resource_json(
        __name__, 'data/T0212_S0039_preferred.json')
    preferred_event_dictionary['labels'] = preferred_event_labels
    if new_event_id == 'T1234':
        new_event_dictionary = dict(
            graceid="T1234",
            instruments="I1,J1,K1,L1,M1",
            group="CBC",
            pipeline="gstlal",
            offline=False,
            superevent="maybe S0039 or empty",
            far=1e-30,
            labels=[],
            extra_attributes=dict(
                CoincInspiral=dict(snr=30.0),
                SingleInspiral=[
                    {'ifo': ifo} for ifo in "I1,J1,K1,L1,M1".split(',')
                ]
            )
        )
    elif new_event_id == 'T0212':
        new_event_dictionary = preferred_event_dictionary.copy()

    new_event_dictionary['labels'] = new_event_labels

    superevent_s0039 = superevents._SuperEvent(
        1163905214.44,
        1163905239.44,
        1163905224.44,
        'S0039',
        preferred_event='T0212',
        event_dict={
            "labels": superevent_labels, "superevent_id": "S0039",
            "submitter": "deep.chatterjee@LIGO.ORG",
            "preferred_event": "T0212",
            "t_start": 1163905214.44,
            "em_events": [],
            "t_0": 1163905224.44,
            "gw_events": ["T0212", "T0211", "T0210"],
            "t_end": 1163905239.44,
            "created": "2018-05-09 07:23:16 UTC"
        }
    )

    with patch('gwcelery.tasks.gracedb.update_superevent') as p, \
            patch('gwcelery.tasks.gracedb.create_label.run') as create_label, \
            patch('gwcelery.tasks.gracedb.get_event',
                  return_value=preferred_event_dictionary):
        superevents._update_superevent(superevent_s0039,
                                       new_event_dictionary,
                                       None,
                                       None,
                                       None)
        # presence of EM_Selected should not
        if 'EM_Selected' in superevent_labels:
            p.assert_not_called()

        else:
            is_complete_preferred = superevents.is_complete(
                preferred_event_dictionary
            )
            is_complete_new = superevents.is_complete(
                new_event_dictionary
            )

            if new_event_id == 'T1234' and (
                    is_complete_preferred and not is_complete_new):
                p.assert_not_called()
                create_label.assert_not_called()
            elif new_event_id == 'T1234' and (
                    not is_complete_preferred and is_complete_new):
                p.assert_called_with('S0039', preferred_event='T1234',
                                     t_0=None)
                create_label.assert_called_once_with('EM_READY', 'S0039')

            elif new_event_id == 'T1234' and (
                    is_complete_preferred and is_complete_new):
                p.assert_called_with('S0039', preferred_event='T1234',
                                     t_0=None)
                create_label.assert_called_once_with('EM_READY', 'S0039')
            elif new_event_id == 'T1234' and (
                    not is_complete_preferred and not is_complete_new):
                p.assert_called_with('S0039', preferred_event='T1234',
                                     t_0=None)
                create_label.assert_called_once_with('EM_READY', 'S0039')
            # this is the case of label addition to existing preferred event
            elif new_event_id == 'T0212':
                p.assert_not_called()
                # label addition completes event
                if not is_complete_preferred and is_complete_new:
                    create_label.assert_called_once_with('EM_READY', 'S0039')
                else:
                    create_label.assert_not_called()


@pytest.mark.parametrize('labels',
                         [['PASTRO_READY', 'RAVEN_ALERT'],
                          ['SKYMAP_READY', 'EMBRIGHT_READY',
                           'PASTRO_READY', 'RAVEN_ALERT']])
@patch('gwcelery.tasks.gracedb.create_label.run')
@patch('gwcelery.tasks.gracedb.get_superevent', s100response)
def test_raven_alert(mock_create_label, labels):
    payload = {
        "uid": "G000003",
        "alert_type": "label_added",
        "description": "",
        "object": {
            "graceid": "G000003",
            "pipeline": "gstlal",
            "group": "CBC",
            "search": "AllSky",
            "far": 1.e-6,
            "instruments": "H1,L1",
            "superevent": "S100",
            "offline": False,
            "gpstime": 1000000,
            "labels": labels,
            'extra_attributes': {
                'SingleInspiral': [
                    {
                        'chisq': 1.07,
                        'snr': 10.95,
                        'ifo': 'L1'
                    },
                    {
                        'chisq': 0.54,
                        'snr': 12.35,
                        'ifo': 'H1'
                    }
                ],
                'CoincInspiral': {
                    'snr': 10.17
                },
            },
        },
        "data": {
            "name": "RAVEN_ALERT"
        }
    }
    superevents.handle(payload)
    calls = [call('ADVREQ', 'S100')]
    if {'SKYMAP_READY', 'EMBRIGHT_READY', 'PASTRO_READY'}.issubset(labels):
        calls.append(call('EM_Selected', 'S100'))
    mock_create_label.assert_has_calls(calls)


@pytest.mark.parametrize('labels',
                         [['EMBRIGHT_READY', 'PASTRO_READY'],
                          ['SKYMAP_READY', 'EMBRIGHT_READY', 'PASTRO_READY']])
def test_is_complete(labels):
    mock_event_dict = _mock_event('G000002')
    mock_event_dict['labels'] = labels
    result = superevents.is_complete(mock_event_dict)
    if len(labels) == 2:
        assert result is False
    elif len(labels) == 3:
        assert result is True


@pytest.mark.parametrize(
    'labels,label',
    [[['EMBRIGHT_READY', 'PASTRO_READY'], 'EMBRIGHT_READY'],
     [['SKYMAP_READY', 'EMBRIGHT_READY', 'PASTRO_READY'], 'SKYMAP_READY']])
def test_process_called(labels, label):
    """Test whether the :meth:`superevents.process` is called
    new type lvalerts, and label additions that complete the event.
    """
    payload = {
        "alert_type": "label_added",
        "data": {"name": label},
        "object": {
            "graceid": "ABCD",
            "far": 1e-6,
            "group": "CBC",
            "labels": labels
        }
    }
    with patch('gwcelery.tasks.superevents.process.run') as process:
        superevents.handle(payload)
        if superevents.is_complete(payload['object']):
            process.assert_called()
        else:
            process.assert_not_called()


def _mock_superevents(*args, **kwargs):
    return [
        {
            "superevent_id": "S123456",
            "preferred_event": "G000002",
            "t_start": 99.0,
            "t_0": 100.0,
            "gw_events": [
                "G000002",
            ],
            "t_end": 101.0,
            "labels": []
        }
    ]


def _mock_event(event):
    if event == "G000002":
        return {
            "graceid": "G000002",
            "gpstime": 100.0,
            "pipeline": "gstlal",
            "group": "CBC",
            "far": 1.e-31,
            "instruments": "H1,L1",
            "labels": [],
            "extra_attributes": {
                "CoincInspiral": {"snr": 20},
                "SingleInspiral": [{"ifo": ifo} for ifo in ["H1", "L1"]]
            },
            "offline": False
        }


@patch('gwcelery.tasks.gracedb.get_superevents', _mock_superevents)
@patch('gwcelery.tasks.gracedb.get_event', _mock_event)
def test_upload_same_event():
    """New event uploaded with the same coinc file as an
    existing event G000002. This could happen during testing
    the superevent manager.
    """
    # payload same as _mock_event except graceid
    payload = {
        "uid": "G000003",
        "alert_type": "new",
        "description": "",
        "object": {
            "graceid": "G000003",
            "gpstime": 100.0,
            "pipeline": "gstlal",
            "group": "CBC",
            "far": 1.e-31,
            "instruments": "H1,L1",
            "extra_attributes": {
                "CoincInspiral": {"snr": 20},
                "SingleInspiral": [{"ifo": ifo} for ifo in ["H1", "L1"]]
            },
            "offline": False,
            "labels": []
        },
        "data": {
            "graceid": "G000003",
            "gpstime": 100.0,
            "pipeline": "gstlal",
            "group": "CBC",
            "far": 1.e-31,
            "instruments": "H1,L1",
            "extra_attributes": {
                "CoincInspiral": {"snr": 20},
                "SingleInspiral": [{"ifo": ifo} for ifo in ["H1", "L1"]]
            },
            "offline": False,
            "labels": []
        }
    }
    with patch('gwcelery.tasks.gracedb.add_event_to_superevent') as p1, \
            patch('gwcelery.tasks.gracedb.update_superevent') as p2:
        superevents.handle(payload)
        p1.assert_called_once()
        p2.assert_not_called()


@pytest.mark.parametrize(
    'group,pipeline,offline,far,instruments,labels,expected_result',
    [['CBC', 'gstlal', False, 1.e-10, 'H1',
        ['PASTRO_READY', 'SKYMAP_READY', 'EMBRIGHT_READY'], True],
     ['CBC', 'gstlal', False, 1.e-10, 'H1',
         ['PASTRO_READY', 'SKYMAP_READY'], False],
     ['CBC', 'gstlal', False, 1.e-6, 'H1,L1', ['EMBRIGHT_READY'], False],
     ['Burst', 'cwb', False, 1e-15, 'H1,L1', ['SKYMAP_READY'], True],
     ['Burst', 'cwb', False, 1e-15, 'H1,L1', [], False],
     ['Burst', 'cwb', True, 1e-30, 'H1,L1,V1', [], False],
     ['CBC', 'gstlal', False, 1.e-6, 'H1,L1,V1', ['RAVEN_ALERT'], False],
     ['CBC', 'gstlal', False, 1.e-6, 'H1,L1,V1', [
      'PASTRO_READY', 'SKYMAP_READY', 'EMBRIGHT_READY', 'RAVEN_ALERT'], True],
     ['CBC', 'gstlal', False, 1.e-15, 'H1,L1,V1', [
      'PASTRO_READY', 'SKYMAP_READY', 'EMBRIGHT_READY', 'RAVEN_ALERT'],
      True]])
def test_should_publish(group, pipeline, offline, far, instruments, labels,
                        expected_result):
    event = dict(graceid='G123456',
                 group=group,
                 pipeline=pipeline,
                 labels=labels,
                 far=far,
                 offline=offline,
                 instruments=instruments)
    result = (superevents.is_complete(event) and
              superevents.should_publish(event))
    assert result == expected_result


def test_parse_trigger_raising():
    garbage_payload = dict(some_value=999, something_else=99)
    with pytest.raises(KeyError):
        superevents.handle(garbage_payload)


response_bad_gateway = Response()
response_bad_gateway.status_code = 502


@patch('gwcelery.tasks.gracedb.create_superevent',
       side_effect=HTTPError(response=response_bad_gateway))
def test_raising_http_error(failing_create_superevent):
    payload = {
        "uid": "G000003",
        "alert_type": "new",
        "description": "",
        "object": {
            "graceid": "G000003",
            "gpstime": 100.0,
            "pipeline": "gstlal",
            "group": "CBC",
            "far": 1.e-31,
            "instruments": "H1,L1",
            "extra_attributes": {
                "CoincInspiral": {"snr": 20}
            },
            "offline": False,
            "labels": []
        },
        "data": {
            "graceid": "G000003",
            "gpstime": 100.0,
            "pipeline": "gstlal",
            "group": "CBC",
            "far": 1.e-31,
            "instruments": "H1,L1",
            "extra_attributes": {
                "CoincInspiral": {"snr": 20}
            },
            "offline": False,
            "labels": []
        }
    }
    with pytest.raises(gracedb.RetryableHTTPError):
        with pytest.raises(exceptions.Retry):
            superevents.handle.delay(payload)


def test_parse_trigger_cbc_1(mock_db):
    """New trigger G000000, less significant than already
    existing superevent. Superevent window much larger than event
    window of G000000.
    """
    # New trigger G000000 time falls in S0039 window
    # addEventToSuperevent should be called
    # updateSuperevent should not be updated
    event_dictionary = {'graceid': 'G000000',
                        'gpstime': 1163905224.4332082,
                        'group': 'CBC',
                        'pipeline': 'gstlal',
                        'offline': False,
                        'far': 3e-09,
                        'instruments': 'H1,L1',
                        'labels': [],
                        'extra_attributes': {
                            'CoincInspiral': {'snr': 10.0},
                            'SingleInspiral': [
                                {'ifo': ifo} for ifo in ['H1', 'L1']]}}
    payload = dict(lvalert_content,
                   object=event_dictionary,
                   data=event_dictionary,
                   alert_type='new',
                   uid='G000000')
    with patch('gwcelery.tasks.gracedb.add_event_to_superevent') as p1, \
            patch('gwcelery.tasks.gracedb.update_superevent') as p2:
        superevents.handle(payload)
        p1.assert_called_once()
        p2.assert_not_called()


def test_parse_trigger_cbc_2(mock_db):
    """New trigger G000003, more significant than already
    existing superevent. Superevent window is much larger that
    event window of G000000
    """
    event_dictionary = {'graceid': 'G000003',
                        'gpstime': 1163905224.4332082,
                        'group': 'CBC',
                        'pipeline': 'gstlal',
                        'offline': False,
                        'far': 3e-31,
                        'instruments': 'H1,L1',
                        'labels': [],
                        'extra_attributes': {
                            'CoincInspiral': {'snr': 30.0},
                            'SingleInspiral': [
                                {"ifo": ifo} for ifo in ["H1", "L1"]]}}
    payload = dict(lvalert_content,
                   object=event_dictionary,
                   data=event_dictionary,
                   alert_type='new',
                   uid='G000003')
    # addEventToSuperevent should be called
    # preferred event should be updated, t_0 should change
    with patch('gwcelery.tasks.gracedb.add_event_to_superevent') as p1, \
            patch('gwcelery.tasks.gracedb.update_superevent') as p2, \
            patch('gwcelery.tasks.gracedb.create_label.run') as create_label:
        superevents.handle(payload)
        p1.assert_called_once()
        p2.assert_called_once_with('S0039', preferred_event='G000003',
                                   t_0=1163905224.4332082)
        create_label.assert_called_once_with('ADVREQ', 'S0039')
        if superevents.is_complete(event_dictionary):
            create_label.assert_called_once_with('EM_READY', 'S0039')
        else:
            assert_not_called_with(create_label, 'EM_READY', 'S0039')


def test_parse_trigger_cbc_3(mock_db):
    """New trigger G000001, not present among superevents
    New superevent created.
    """
    event_dictionary = {'graceid': 'G000001',
                        'gpstime': 1286741861.52678,
                        'group': 'CBC',
                        'pipeline': 'gstlal',
                        'offline': False,
                        'far': 3e-31,
                        'instruments': 'H1,L1,V1',
                        'labels': [],
                        'extra_attributes': {
                            'CoincInspiral': {'snr': 12.0},
                            'SingleInspiral': [
                                {"ifo": ifo} for ifo in
                                ["H1", "L1", "V1"]]}}
    payload = dict(lvalert_content,
                   object=event_dictionary,
                   data=event_dictionary,
                   alert_type='new',
                   uid='G000001')
    # G000001 absent in any superevent window, new superevent created
    # createSuperevent should be called exactly once
    with patch('gwcelery.tasks.gracedb.create_superevent') as p, \
            patch('gwcelery.tasks.gracedb.create_label'):
        superevents.handle(payload)
        p.assert_called_once()


def test_parse_trigger_cbc_4(mock_db):
    """New trigger G000002, doesn't pass far threshold"""
    event_dictionary = {'graceid': 'G000002',
                        'gpstime': 1286741861.52678,
                        'group': 'CBC',
                        'pipeline': 'gstlal',
                        'offline': False,
                        'far': 5.5e-02,
                        'instruments': 'H1,L1,V1',
                        'labels': [],
                        'extra_attributes': {
                            'CoincInspiral': {'snr': 4.0},
                            'SingleInspiral': [
                                {"ifo": ifo} for ifo in
                                ["H1", "L1", "V1"]]}}
    payload = dict(lvalert_content,
                   alert_type='new',
                   object=event_dictionary,
                   data=event_dictionary,
                   uid='G000002')
    superevents.handle(payload)
    # neither method is called due to low far
    with patch.object(superevents, 'process') as mock_process:
        superevents.handle(payload)
        mock_process.assert_not_called()


def test_parse_trigger_burst_1(mock_db):
    """New cwb trigger G000005 with gpstime lying partially in
    S0039 window, not more significant than existing preferred
    event, superevent window changed.
    """
    event_dictionary = {'graceid': 'G000005',
                        'gpstime': 1163905214.4,
                        'group': 'Burst',
                        'pipeline': 'cwb',
                        'offline': False,
                        'far': 3.02e-09,
                        'instruments': 'H1,L1',
                        'labels': [],
                        'extra_attributes': {
                            'MultiBurst': {
                                'duration': 0.23639,
                                'start_time': 1163905215,
                                'snr': 10.3440}}}
    payload = dict(lvalert_content,
                   object=event_dictionary,
                   data=event_dictionary,
                   alert_type='new',
                   uid='G000005')
    # preferred event should not be updated
    # superevent window should change
    with patch('gwcelery.tasks.gracedb.add_event_to_superevent') as p1, \
            patch('gwcelery.tasks.gracedb.update_superevent') as p2:
        superevents.handle(payload)
        p1.assert_called_once()
        p2.assert_called_once_with('S0039',
                                   t_end=pytest.approx(1163905239, abs=1),
                                   t_start=pytest.approx(1163905214, abs=1))


def test_parse_trigger_burst_2(mock_db):
    """New oLIB trigger G000006 with gpstime lying partially in
    S0039 window, more significant than already existing preferred
    event. superevent window changed
    """
    # preferred event not be updated, in spite of lower far of new event
    # because existing preferred event is CBC
    event_dictionary = {'graceid': 'G000006',
                        'gpstime': 1163905239.5,
                        'group': 'Burst',
                        'pipeline': 'oLIB',
                        'offline': False,
                        'far': 3.02e-16,
                        'instruments': 'H1,L1',
                        'labels': [],
                        'extra_attributes': {
                            'LalInferenceBurst': {
                                'quality_mean': 20.6458,
                                'frequency_mean': 117.9644,
                                'omicron_snr_network': 8.77}}}
    payload = dict(lvalert_content,
                   object=event_dictionary,
                   data=event_dictionary,
                   alert_type='new',
                   uid='G000006')
    with patch('gwcelery.tasks.gracedb.add_event_to_superevent') as p1, \
            patch('gwcelery.tasks.gracedb.update_superevent') as p2:
        superevents.handle(payload)
        p1.assert_called_once()
        p2.assert_called_once_with('S0039',
                                   t_end=pytest.approx(1163905239, abs=1),
                                   t_start=pytest.approx(1163905214, abs=1))


def test_parse_trigger_burst_3(mock_db):
    """New oLIB trigger G000007, not present among superevents
    New superevent created.
    Q_mean = frequency_mean = 100., hence d_t_start = d_t_end = 1s
    """
    event_dictionary = {'graceid': 'G000007',
                        'gpstime': 1163905249.5,
                        'group': 'Burst',
                        'pipeline': 'oLIB',
                        'offline': False,
                        'far': 3.02e-16,
                        'instruments': 'H1,L1',
                        'labels': [],
                        'extra_attributes': {
                            'LalInferenceBurst': {
                                'quality_mean': 100.0,
                                'frequency_mean': 100.0,
                                'omicron_snr_network': 8.0}}}
    payload = dict(lvalert_content,
                   object=event_dictionary,
                   data=event_dictionary,
                   alert_type='new',
                   uid='G000007')
    # G000007 absent in any superevent window, new superevent created
    with patch('gwcelery.tasks.gracedb.create_superevent') as p, \
            patch('gwcelery.tasks.gracedb.create_label'):
        superevents.handle(payload)
        p.assert_called_once_with('G000007',
                                  pytest.approx(1163905249.5),
                                  pytest.approx(1163905248.5),
                                  pytest.approx(1163905250.5))


def test_parse_trigger_burst_4(mock_db):
    """New CWB trigger G000008, not present among superevents
    New superevent created.
    extra attribute duration = 0.02s
    """
    event_dictionary = {'graceid': 'G000008',
                        'gpstime': 1128658942.9878,
                        'group': 'Burst',
                        'pipeline': 'CWB',
                        'offline': False,
                        'far': 1.23e-09,
                        'instruments': 'H1,L1',
                        'labels': [],
                        'extra_attributes': {
                            'MultiBurst': {
                                'duration': 0.02,
                                'start_time': 1128658942,
                                'snr': 19.824}}}
    payload = dict(lvalert_content,
                   object=event_dictionary,
                   data=event_dictionary,
                   alert_type='new',
                   uid='G000008')
    # G000008 absent in any superevent window, new superevent created
    with patch('gwcelery.tasks.gracedb.create_superevent') as p, \
            patch('gwcelery.tasks.gracedb.create_label'):
        superevents.handle(payload)
        p.assert_called_once()


def test_S190421ar_spiir_scenario(mock_db):    # noqa: N802
    """Test to ensure that a low FAR event with accidental high
    SNR is not promoted to the preferred event status. For example, here,
    the new event G330298 has SNR 10.51, higher than the preferred event
    G330308 which has SNR 10.17. But the preferred event is not changed on
    the basis of low FAR.
    """
    event_dictionary = {'graceid': 'G330298',
                        'gpstime': 1239917954.40918,
                        'far': 5.57979637960671e-06,
                        'group': 'CBC',
                        'instruments': 'H1,L1',
                        'pipeline': 'spiir',
                        'offline': False,
                        'labels': [],
                        'extra_attributes': {
                            'CoincInspiral': {
                                'snr': 10.5107507705688},
                            'SingleInspiral': [
                                {"ifo": ifo} for ifo in ["H1", "L1"]]}}
    payload = dict(lvalert_content,
                   alert_type='new',
                   object=event_dictionary,
                   data=event_dictionary,
                   uid='G330298')
    with patch('gwcelery.tasks.gracedb.add_event_to_superevent') as p1, \
            patch('gwcelery.tasks.gracedb.update_superevent') as p2:
        superevents.handle(payload)
        p1.assert_called_once()
        p2.assert_not_called()



"""Source Parameter Estimation with LALInference and Bilby."""
from distutils.spawn import find_executable
from distutils.dir_util import mkpath
import glob
import itertools
import json
import os
import shutil
import subprocess
import tempfile
import urllib

from celery import group
from gwdatafind import find_urls
import numpy as np
from requests.exceptions import HTTPError

from .. import app
from ..jinja import env
from .core import ordered_group
from . import condor
from . import gracedb


ini_name = 'online_lalinference_pe.ini'

executables = {'datafind': 'gw_data_find',
               'mergeNSscript': 'lalinference_nest2pos',
               'mergeMCMCscript': 'cbcBayesMCMC2pos',
               'combinePTMCMCh5script': 'cbcBayesCombinePTMCMCh5s',
               'resultspage': 'cbcBayesPostProc',
               'segfind': 'ligolw_segment_query',
               'ligolw_print': 'ligolw_print',
               'coherencetest': 'lalinference_coherence_test',
               'lalinferencenest': 'lalinference_nest',
               'lalinferencemcmc': 'lalinference_mcmc',
               'lalinferencebambi': 'lalinference_bambi',
               'lalinferencedatadump': 'lalinference_datadump',
               'ligo-skymap-from-samples': 'true',
               'ligo-skymap-plot': 'true',
               'processareas': 'process_areas',
               'computeroqweights': 'lalinference_compute_roq_weights',
               'mpiwrapper': 'lalinference_mpi_wrapper',
               'gracedb': 'gracedb',
               'ppanalysis': 'cbcBayesPPAnalysis',
               'pos_to_sim_inspiral': 'cbcBayesPosToSimInspiral',
               'bayeswave': 'BayesWave',
               'bayeswavepost': 'BayesWavePost'}


def _data_exists(end, frametype_dict):
    """Check whether data at end time can be found with gwdatafind and return
    true it it is found.
    """
    return min(
        len(
            find_urls(ifo[0], frametype_dict[ifo], end, end + 1)
        ) for ifo in frametype_dict.keys()
    ) > 0


class NotEnoughData(Exception):
    """Raised if found data is not enough due to the latency of data
    transfer
    """


@app.task(bind=True, autoretry_for=(NotEnoughData, ), default_retry_delay=1,
          max_retries=86400, retry_backoff=True, shared=False)
def query_data(self, trigtime):
    """Continues to query data until it is found with gwdatafind and return
    frametypes for the data. If data is not found in 86400 seconds = 1 day,
    raise NotEnoughData.
    """
    end = trigtime + 2
    if _data_exists(end, app.conf['low_latency_frame_types']):
        return app.conf['low_latency_frame_types']
    elif _data_exists(end, app.conf['high_latency_frame_types']):
        return app.conf['high_latency_frame_types']
    else:
        raise NotEnoughData


@app.task(ignore_result=True, shared=False)
def upload_no_frame_files(request, exc, traceback, superevent_id):
    """Upload notification when no frame files are found.

    Parameters
    ----------
    request : Context (placeholder)
        Task request variables
    exc : Exception
        Exception rased by condor.submit
    traceback : str (placeholder)
        Traceback message from a task
    superevent_id : str
        The GraceDB ID of a target superevent

    """
    if isinstance(exc, NotEnoughData):
        gracedb.upload.delay(
            filecontents=None, filename=None,
            graceid=superevent_id,
            message='Frame files have not been found.',
            tags='pe'
        )


def _find_appropriate_cal_env(trigtime, dir_name):
    """Return the path to the calibration uncertainties estimated at the time
    before and closest to the trigger time. If there are no calibration
    uncertainties estimated before the trigger time, return the oldest one. The
    gpstimes at which the calibration uncertainties were estimated and the
    names of the files containing the uncertaintes are saved in
    [HLV]_CalEnvs.txt.

    Parameters
    ----------
    trigtime : float
        The trigger time of a target event
    dir_name : str
        The path to the directory where files containing calibration
        uncertainties exist

    Return
    ------
    path : str
        The path to the calibration uncertainties appropriate for a target
        event

    """
    filename, = glob.glob(os.path.join(dir_name, '[HLV]_CalEnvs.txt'))
    calibration_index = np.atleast_1d(
        np.recfromtxt(filename, names=['gpstime', 'filename'])
    )
    gpstimes = calibration_index['gpstime']
    candidate_gpstimes = gpstimes < trigtime
    if np.any(candidate_gpstimes):
        idx = np.argmax(gpstimes * candidate_gpstimes)
        appropriate_cal = calibration_index['filename'][idx]
    else:
        appropriate_cal = calibration_index['filename'][np.argmin(gpstimes)]
    return os.path.join(dir_name, appropriate_cal.decode('utf-8'))


@app.task(shared=False)
def prepare_ini(frametype_dict, event, superevent_id=None):
    """Determine an appropriate PE settings for the target event and return ini
    file content for LALInference pipeline
    """
    # Get template of .ini file
    ini_template = env.get_template('online_pe.jinja2')

    # fill out the ini template and return the resultant content
    singleinspiraltable = event['extra_attributes']['SingleInspiral']
    trigtime = event['gpstime']
    ini_settings = {
        'gracedb_host': app.conf['gracedb_host'],
        'types': frametype_dict,
        'channels': app.conf['strain_channel_names'],
        'state_vector_channels': app.conf['state_vector_channel_names'],
        'webdir': os.path.join(
            app.conf['pe_results_path'], event['graceid'], 'lalinference'
        ),
        'paths': [{'name': name, 'path': find_executable(executable)}
                  for name, executable in executables.items()],
        'h1_calibration': _find_appropriate_cal_env(
            trigtime,
            '/home/cbc/pe/O3/calibrationenvelopes/LIGO_Hanford'
        ),
        'l1_calibration': _find_appropriate_cal_env(
            trigtime,
            '/home/cbc/pe/O3/calibrationenvelopes/LIGO_Livingston'
        ),
        'v1_calibration': _find_appropriate_cal_env(
            trigtime,
            '/home/cbc/pe/O3/calibrationenvelopes/Virgo'
        ),
        'mc': min([sngl['mchirp'] for sngl in singleinspiraltable]),
        'q': min([sngl['mass2'] / sngl['mass1']
                  for sngl in singleinspiraltable]),
        'mpirun': find_executable('mpirun')
    }
    ini_rota = ini_template.render(ini_settings)
    ini_settings.update({'use_of_ini': 'online'})
    ini_online = ini_template.render(ini_settings)
    # upload LALInference ini file to GraceDB
    if superevent_id is not None:
        gracedb.upload.delay(
            ini_rota, filename=ini_name, graceid=superevent_id,
            message=('Automatically generated LALInference configuration file'
                     ' for this event.'),
            tags='pe')

    return ini_online


def pre_pe_tasks(event, superevent_id):
    """Return canvas of tasks executed before parameter estimation starts"""
    return query_data.s(event['gpstime']).on_error(
        upload_no_frame_files.s(superevent_id)
    ) | prepare_ini.s(event, superevent_id)


@app.task(shared=False)
def _setup_dag_for_lalinference(coinc_psd, ini_contents,
                                rundir, superevent_id):
    """Create DAG for a lalinference run and return the path to DAG.

    Parameters
    ----------
    coinc_psd : tuple of byte contents
        Tuple of the byte contents of ``coinc.xml`` and ``psd.xml.gz``
    ini_contents : str
        The content of online_lalinference_pe.ini
    rundir : str
        The path to a run directory where the DAG file exits
    superevent_id : str
        The GraceDB ID of a target superevent

    Returns
    -------
    path_to_dag : str
        The path to the .dag file

    """
    coinc_contents, psd_contents = coinc_psd

    # write down coinc.xml in the run directory
    path_to_coinc = os.path.join(rundir, 'coinc.xml')
    with open(path_to_coinc, 'wb') as f:
        f.write(coinc_contents)

    # write down psd.xml.gz
    if psd_contents is not None:
        path_to_psd = os.path.join(rundir, 'psd.xml.gz')
        with open(path_to_psd, 'wb') as f:
            f.write(psd_contents)
        psd_arg = ['--psd', path_to_psd]
    else:
        psd_arg = []

    # write down .ini file in the run directory.
    path_to_ini = os.path.join(rundir, ini_name)
    with open(path_to_ini, 'w') as f:
        f.write(ini_contents)

    try:
        subprocess.run(
            ['lalinference_pipe', '--run-path', rundir,
             '--coinc', path_to_coinc, path_to_ini] + psd_arg,
            capture_output=True, check=True)
    except subprocess.CalledProcessError as e:
        contents = b'args:\n' + json.dumps(e.args[1]).encode('utf-8') + \
                   b'\n\nstdout:\n' + e.stdout + b'\n\nstderr:\n' + e.stderr
        gracedb.upload.delay(
            filecontents=contents, filename='pe_dag.log',
            graceid=superevent_id,
            message='Failed to prepare DAG for lalinference', tags='pe'
        )
        shutil.rmtree(rundir)
        raise
    else:
        # Remove the ini file so that people do not accidentally use this ini
        # file and hence online-PE-only nodes.
        os.remove(path_to_ini)

    return os.path.join(rundir, 'multidag.dag')


@app.task(shared=False)
def _setup_dag_for_bilby(event, rundir, preferred_event_id, superevent_id):
    """Create DAG for a bilby run and return the path to DAG.

    Parameters
    ----------
    event : json contents
        The json contents retrieved from gracedb.get_event()
    rundir : str
        The path to a run directory where the DAG file exits
    preferred_event_id : str
        The GraceDB ID of a target preferred event
    superevent_id : str
        The GraceDB ID of a target superevent

    Returns
    -------
    path_to_dag : str
        The path to the .dag file

    """
    path_to_json = os.path.join(rundir, 'event.json')
    with open(path_to_json, 'w') as f:
        json.dump(event, f, indent=2)

    path_to_webdir = os.path.join(
        app.conf['pe_results_path'], preferred_event_id, 'bilby'
    )

    setup_arg = ['bilby_pipe_gracedb', '--webdir', path_to_webdir,
                 '--outdir', rundir, '--json', path_to_json, '--online-pe',
                 '--convert-to-flat-in-component-mass']

    if not app.conf['gracedb_host'] == 'gracedb.ligo.org':
        setup_arg += ['--channel-dict', 'o2replay',
                      '--sampler-kwargs', 'FastTest']
    try:
        subprocess.run(setup_arg, capture_output=True, check=True)
    except subprocess.CalledProcessError as e:
        contents = b'args:\n' + json.dumps(e.args[1]).encode('utf-8') + \
                   b'\n\nstdout:\n' + e.stdout + b'\n\nstderr:\n' + e.stderr
        gracedb.upload.delay(
            filecontents=contents, filename='pe_dag.log',
            graceid=superevent_id,
            message='Failed to prepare DAG for bilby', tags='pe'
        )
        shutil.rmtree(rundir)
        raise
    else:
        # Uploads bilby ini file to GraceDB
        group(upload_results_tasks(
            rundir, 'bilby_config.ini', superevent_id,
            'Automatically generated Bilby configuration file',
            'pe', 'online_bilby_pe.ini')).delay()

    path_to_dag, = glob.glob(os.path.join(rundir, 'submit/dag*.submit'))
    print(path_to_dag)
    return path_to_dag


@app.task(shared=False)
def _condor_no_submit(path_to_dag):
    """Run 'condor_submit_dag -no_submit' and return the path to .sub file."""
    subprocess.run(['condor_submit_dag', '-no_submit', path_to_dag],
                   capture_output=True, check=True)
    return '{}.condor.sub'.format(path_to_dag)


@app.task(shared=False)
def dag_prepare_task(rundir, superevent_id, preferred_event_id, pe_pipeline,
                     ini_contents=None):
    """Return a canvas of tasks to prepare DAG.

    Parameters
    ----------
    rundir : str
        The path to a run directory where the DAG file exits
    superevent_id : str
        The GraceDB ID of a target superevent
    preferred_event_id : str
        The GraceDB ID of a target preferred event
    pe_pipeline : str
        The parameter estimation pipeline used
        Either 'lalinference' OR 'bilby'
    ini_contents : str
        The content of online_lalinference_pe.ini
        Required if pe_pipeline == 'lalinference'

    Returns
    -------
    canvas : canvas of tasks
        The canvas of tasks to prepare DAG

    """
    if pe_pipeline == 'lalinference':
        canvas = ordered_group(
            gracedb.download.si('coinc.xml', preferred_event_id),
            _download_psd.si(preferred_event_id)
        ) | _setup_dag_for_lalinference.s(ini_contents, rundir, superevent_id)
    elif pe_pipeline == 'bilby':
        canvas = gracedb.get_event.si(preferred_event_id) | \
            _setup_dag_for_bilby.s(rundir, preferred_event_id, superevent_id)
    else:
        raise NotImplementedError(f'Unknown PE pipeline {pe_pipeline}.')
    canvas |= _condor_no_submit.s()
    return canvas


def _find_paths_from_name(directory, name):
    """Return the paths of files or directories with given name under the
    specfied directory

    Parameters
    ----------
    directory : string
        Name of directory under which the target file or directory is searched
        for.
    name : string
        Name of target files or directories

    Returns
    -------
    paths : generator
        Paths to the target files or directories

    """
    return glob.iglob(os.path.join(directory, '**', name), recursive=True)


@app.task(ignore_result=True, shared=False)
def job_error_notification(request, exc, traceback,
                           superevent_id, rundir, pe_pipeline):
    """Upload notification when condor.submit terminates unexpectedly.

    Parameters
    ----------
    request : Context (placeholder)
        Task request variables
    exc : Exception
        Exception rased by condor.submit
    traceback : str (placeholder)
        Traceback message from a task
    superevent_id : str
        The GraceDB ID of a target superevent
    rundir : str
        The run directory for PE
    pe_pipeline : str
        The parameter estimation pipeline used
        Either lalinference OR bilby

    """
    if isinstance(exc, condor.JobAborted):
        gracedb.upload.delay(
            filecontents=None, filename=None, graceid=superevent_id, tags='pe',
            message='The {} condor job was aborted.'.format(pe_pipeline)
        )
    elif isinstance(exc, condor.JobFailed):
        gracedb.upload.delay(
            filecontents=None, filename=None, graceid=superevent_id, tags='pe',
            message='The {} condor job failed.'.format(pe_pipeline)
        )
    # Get paths to .log files, .err files, .out files
    paths_to_log = _find_paths_from_name(rundir, '*.log')
    paths_to_err = _find_paths_from_name(rundir, '*.err')
    paths_to_out = _find_paths_from_name(rundir, '*.out')
    # Upload .log and .err files
    for path in itertools.chain(paths_to_log, paths_to_err, paths_to_out):
        with open(path, 'rb') as f:
            contents = f.read()
        if contents:
            # put .log suffix in log file names so that users can directly
            # read the contents instead of downloading them when they click
            # file names
            gracedb.upload.delay(
                filecontents=contents,
                filename=os.path.basename(path) + '.log',
                graceid=superevent_id,
                message='A log file for {} condor job.'.format(pe_pipeline),
                tags='pe'
            )


@app.task(ignore_result=True, shared=False)
def _upload_url(pe_results_path, graceid, pe_pipeline):
    """Upload url of a page containing all of the plots."""
    if pe_pipeline == 'lalinference':
        path_to_posplots, = _find_paths_from_name(
            pe_results_path, 'posplots.html'
        )
    elif pe_pipeline == 'bilby':
        path_to_posplots, = _find_paths_from_name(
            pe_results_path, 'home.html'
        )
    else:
        raise NotImplementedError(f'Unknown PE pipeline {pe_pipeline}.')

    baseurl = urllib.parse.urljoin(
                  app.conf['pe_results_url'],
                  os.path.relpath(
                      path_to_posplots,
                      app.conf['pe_results_path']
                  )
              )
    gracedb.upload.delay(
        filecontents=None, filename=None, graceid=graceid,
        message=('Online {} parameter estimation finished.'
                 '<a href={}>results</a>').format(pe_pipeline, baseurl),
        tags='pe'
    )


def upload_results_tasks(pe_results_path, filename, graceid, message, tag,
                         uploaded_filename=None):
    """Return tasks to get the contents of PE result files and upload them to
    GraceDB.

    Parameters
    ----------
    pe_results_path : string
        Directory under which the target file located.
    filename : string
        Name of the target file
    graceid : string
        GraceDB ID
    message : string
        Message uploaded to GraceDB
    tag : str
        Name of tag to add the GraceDB log
    uploaded_filename : str
        Name of the uploaded file. If not supplied, it is the same as the
        original file name.

    Returns
    -------
    tasks : list of celery tasks

    """
    tasks = []
    for path in _find_paths_from_name(pe_results_path, filename):
        if uploaded_filename is None:
            _uploaded_filename = os.path.basename(path)
        else:
            _uploaded_filename = uploaded_filename
        with open(path, 'rb') as f:
            tasks.append(gracedb.upload.si(f.read(), _uploaded_filename,
                                           graceid, message, tag))
    return tasks


@app.task(ignore_result=True, shared=False)
def clean_up(rundir):
    """Clean up a run directory.

    Parameters
    ----------
    rundir : str
        The path to a run directory where the DAG file exits

    """
    shutil.rmtree(rundir)


@app.task(ignore_result=True, shared=False)
def dag_finished(rundir, preferred_event_id, superevent_id, pe_pipeline):
    """Upload PE results and clean up run directory

    Parameters
    ----------
    rundir : str
        The path to a run directory where the DAG file exits
    preferred_event_id : str
        The GraceDB ID of a target preferred event
    superevent_id : str
        The GraceDB ID of a target superevent
    pe_pipeline : str
        The parameter estimation pipeline used
        Either lalinference OR bilby

    Returns
    -------
    tasks : canvas
        The work-flow for uploading PE results

    """
    pe_results_path = os.path.join(
        app.conf['pe_results_path'], preferred_event_id, pe_pipeline
    )

    if pe_pipeline == 'lalinference':
        uploads = [
            (rundir, 'glitch_median_PSD_forLI_*.dat',
             'Bayeswave PSD used for LALInference PE', None),
            (rundir, 'lalinference*.dag', 'LALInference DAG', None),
            (rundir, 'posterior*.hdf5',
             'LALInference posterior samples',
             'LALInference.posterior_samples.hdf5'),
            (pe_results_path, 'extrinsic.png',
             'LALInference corner plot for extrinsic parameters',
             'LALInference.extrinsic.png'),
            (pe_results_path, 'sourceFrame.png',
             'LALInference corner plot for source frame parameters',
             'LALInference.intrinsic.png')
        ]
    elif pe_pipeline == 'bilby':
        resultdir = os.path.join(rundir, 'result')
        uploads = [
            (resultdir, '*merge_result.json',
             'Bilby posterior samples',
             'Bilby.posterior_samples.json'),
            (resultdir, '*_extrinsic_corner.png',
             'Bilby corner plot for extrinsic parameters',
             'Bilby.extrinsic.png'),
            (resultdir, '*_intrinsic_corner.png',
             'Bilby corner plot for intrinsic parameters',
             'Bilby.intrinsic.png')
        ]
    else:
        raise NotImplementedError(f'Unknown PE pipeline {pe_pipeline}.')

    upload_tasks = []
    for dir, name1, comment, name2 in uploads:
        upload_tasks += upload_results_tasks(
            dir, name1, superevent_id, comment, 'pe', name2)

    # FIXME: _upload_url.si has to be out of group for
    # gracedb.create_label.si to run
    (
        _upload_url.si(pe_results_path, superevent_id, pe_pipeline)
        |
        group(upload_tasks)
        |
        clean_up.si(rundir)
    ).delay()

    if pe_pipeline == 'lalinference':
        gracedb.create_label.delay('PE_READY', superevent_id)


@gracedb.task(shared=False)
def _download_psd(gid):
    """Download ``psd.xml.gz`` and return its content. If that file does not
    exist, return None.
    """
    try:
        return gracedb.download("psd.xml.gz", gid)
    except HTTPError:
        return None


@app.task(ignore_result=True, shared=False)
def start_pe(ini_contents, preferred_event_id, superevent_id, pe_pipeline):
    """Run Parameter Estimation on a given event.

    Parameters
    ----------
    ini_contents : str
        The content of online_lalinference_pe.ini
    preferred_event_id : str
        The GraceDB ID of a target preferred event
    superevent_id : str
        The GraceDB ID of a target superevent
    pe_pipeline : str
        The parameter estimation pipeline used
        lalinference OR bilby

    """
    gracedb.upload.delay(
        filecontents=None, filename=None, graceid=superevent_id,
        message=('Starting {} online parameter estimation '
                 'for {}').format(pe_pipeline, preferred_event_id),
        tags='pe'
    )

    # make a run directory
    pipeline_dir = os.path.expanduser('~/.cache/{}'.format(pe_pipeline))
    mkpath(pipeline_dir)
    rundir = tempfile.mkdtemp(
        dir=pipeline_dir, prefix='{}_'.format(superevent_id)
    )

    # give permissions to read the files under the run directory so that PE
    # ROTA people can check the status of parameter estimation.
    os.chmod(rundir, 0o755)

    canvas = (
        dag_prepare_task(
            rundir, superevent_id, preferred_event_id,
            pe_pipeline, ini_contents
        )
        |
        condor.submit.s().on_error(
            job_error_notification.s(superevent_id, rundir, pe_pipeline)
        )
        |
        dag_finished.si(
            rundir, preferred_event_id, superevent_id, pe_pipeline
        )
    )
    canvas.delay()



"""Package data helpers."""
import json
import pickle

import pkg_resources

__all__ = ('resource_json', 'resource_pickle')


def resource_json(*args, **kwargs):
    """Load a JSON file from package data."""
    with pkg_resources.resource_stream(*args, **kwargs) as f:
        return json.load(f)


def resource_pickle(*args, **kwargs):
    """Load a JSON file from package data."""
    with pkg_resources.resource_stream(*args, **kwargs) as f:
        return pickle.load(f)



from subprocess import CalledProcessError

from sentry_sdk.integrations import Integration
from sentry_sdk.scope import add_global_event_processor


class SubprocessIntegration(Integration):
    """Capture stderr and stdout from CalledProcessError exceptions."""

    identifier = 'subprocess'

    @staticmethod
    def setup_once():

        @add_global_event_processor
        def capture(event, hint):
            if 'exc_info' not in hint:
                return event

            _, e, _ = hint['exc_info']
            if not isinstance(e, CalledProcessError):
                return event

            breadcrumbs = event.get('breadcrumbs')
            if not breadcrumbs:
                return event

            data = breadcrumbs[-1].setdefault('data', {})
            for key in ['stderr', 'stdout']:
                value = getattr(e, key)
                data[key] = value.decode(errors='replace')

            return event



import json
from unittest.mock import patch

import h5py
import numpy as np
import pytest

from ..tasks import em_bright
from ..util.tempfile import NamedTemporaryFile


@patch('gwcelery.tasks.gracedb.download.run', return_value='mock_em_bright')
@patch('gwcelery.tasks.gracedb.upload.run')
@patch('gwcelery.tasks.em_bright.plot.run')
def test_handle_em_bright_json(mock_plot, mock_upload, mock_download):
    alert = {
        "data": {
            "tag_names": ["em_bright"],
            "filename": "em_bright.json"
        },
        "uid": "TS123456",
        "alert_type": "log"
    }
    em_bright.handle(alert)
    mock_download.assert_called_once_with(
        'em_bright.json',
        'TS123456'
    )
    mock_plot.assert_called_once_with('mock_em_bright')
    mock_upload.assert_called_once()


def test_classify_gstlal():
    res = json.loads(em_bright.classifier_gstlal(
        1.355607, 1.279483, 0.0, 0.0, 15.6178))
    assert res['HasNS'] == pytest.approx(1.0, abs=1e-3)
    assert res['HasRemnant'] == pytest.approx(1.0, abs=1e-3)


@pytest.mark.parametrize(
        'posterior_samples,embright',
        [[[(1.2, 1.0, 0.0, 0.0, 100.0),
           (2.0, 0.5, 0.99, 0.99, 150.0)],
          {'HasNS': 1.0, 'HasRemnant': 1.0}],
         [[(20., 12.0, 0.0, 0.0, 200.0),
           (22.0, 11.5, 0.80, 0.00, 250.0),
           (21.0, 10.0, 0.0, 0.0, 250)],
          {'HasNS': 0.0, 'HasRemnant': 0.0}]])
def test_posterior_samples(posterior_samples, embright):
    with NamedTemporaryFile() as f:
        filename = f.name
        with h5py.File(f, 'r+') as tmp_h5:
            data = np.array(
                    posterior_samples,
                    dtype=[('mc', '<f8'), ('q', '<f8'),
                           ('a1', '<f8'), ('a2', '<f8'), ('dist', '<f8')])
            tmp_h5.create_dataset(
                'lalinference/lalinference_mcmc/posterior_samples',
                data=data)
        content = open(filename, 'rb').read()
    r = json.loads(em_bright.em_bright_posterior_samples(content))
    assert r == embright


@pytest.mark.parametrize(
    'args,has_ns,has_remnant',
    [[(1.355607, 1.279483, 0.0, 0.0, 15.6178), 1.0, 1.0],
     [(40.0, 3.0001, 0.0, 0.0, 15), 0.0, 0.0],
     [(40.0, 2.9999, 0.0, 0.0, 15), 1.0, 0.0]])
def test_classify_other(args, has_ns, has_remnant):
    res = json.loads(em_bright.classifier_other(*args))
    assert res['HasNS'] == has_ns
    assert res['HasRemnant'] == has_remnant



"""Qualitative source properties for CBC events."""
import io
import json
from matplotlib import pyplot as plt

from ligo import computeDiskMass, em_bright

from celery.utils.log import get_task_logger

from ..import app
from . import gracedb, lvalert
from .p_astro import _format_prob
from ..util import NamedTemporaryFile, PromiseProxy, resource_pickle

NS_CLASSIFIER = PromiseProxy(
    resource_pickle, ('ligo.data', 'knn_ns_classifier.pkl'))
EM_CLASSIFIER = PromiseProxy(
    resource_pickle, ('ligo.data', 'knn_em_classifier.pkl'))

log = get_task_logger(__name__)


@lvalert.handler('superevent',
                 'mdc_superevent',
                 shared=False)
def handle(alert):
    """LVAlert handler to plot and upload a visualization
    of every ``em_bright.json``.
    """
    filename = 'em_bright.json'
    graceid = alert['uid']
    if alert['alert_type'] == 'log' and alert['data']['filename'] == filename:
        (
            gracedb.download.si(filename, graceid)
            |
            plot.s()
            |
            gracedb.upload.s(
                filename.replace('.json', '.png'),
                graceid,
                message=(
                    'Source properties visualization from '
                    '<a href="/api/superevents/{graceid}/files/{filename}">'
                    '{filename}</a>').format(
                        graceid=graceid, filename=filename),
                tags=['em_follow', 'em_bright', 'public']
            )
        ).delay()


@app.task(shared=False)
def plot(contents):
    """Make a visualization of the source properties.

    Examples
    --------
    .. plot::
        :include-source:

        >>> from gwcelery.tasks import em_bright
        >>> contents = '{"HasNS": 0.9137, "HasRemnant": 0.0}'
        >>> em_bright.plot(contents)
    """
    properties = json.loads(contents)
    outfile = io.BytesIO()

    properties = dict(sorted(properties.items(), reverse=True))
    probs, names = list(properties.values()), list(properties.keys())

    with plt.style.context('seaborn-white'):
        fig, ax = plt.subplots(figsize=(3, 1))
        ax.barh(names, probs)
        ax.barh(names, [1.0 - p for p in probs],
                color='lightgray', left=probs)
        for i, prob in enumerate(probs):
            ax.annotate(_format_prob(prob), (0, i), (4, 0),
                        textcoords='offset points', ha='left', va='center')
        ax.set_xlim(0, 1)
        ax.set_xticks([])
        ax.tick_params(left=False)
        for side in ['top', 'bottom', 'right']:
            ax.spines[side].set_visible(False)
        fig.tight_layout()
        fig.savefig(outfile, format='png')
    return outfile.getvalue()


@app.task(shared=False)
def em_bright_posterior_samples(posterior_file_content):
    """Returns the probability of having a NS component and remnant
    using LALInference posterior samples.

    Parameters
    ----------
    posterior_file_content : hdf5 posterior file content

    Returns
    -------
    str
        JSON formatted string storing ``HasNS`` and ``HasRemnant``
        probabilities

    Examples
    --------
    >>> em_bright_posterior_samples(GraceDb().files('S190930s',
    ... 'LALInference.posterior_samples.hdf5').read())
    {"HasNS": 0.014904901243599122, "HasRemnant": 0.0}

    """
    with NamedTemporaryFile(content=posterior_file_content) as samplefile:
        filename = samplefile.name
        has_ns, has_remnant = em_bright.source_classification_pe(filename)
    data = json.dumps({
        'HasNS': has_ns,
        'HasRemnant': has_remnant
    })
    return data


def _em_bright(m1, m2, c1, c2, threshold=3.0):
    """This is the place-holder function for the source classfication pipeline.
    This placeholder code will only act upon the mass2 point estimate value and
    classify the systems as whether they have a neutron or not.
    """
    disk_mass = computeDiskMass.computeDiskMass(m1, m2, c1, c2)
    p_ns = 1.0 if m2 <= threshold else 0.0
    p_emb = 1.0 if disk_mass > 0.0 or m1 < threshold else 0.0
    return p_ns, p_emb


@app.task(shared=False)
def classifier_other(mass1, mass2, spin1z, spin2z, snr):
    """Returns the boolean probability of having a NS component and the
    probability of having non-zero disk mass. This method is used for pipelines
    that do not provide the data products necessary for computation of the
    source properties probabilities.

    Parameters
    ----------
    mass1 : float
        Primary mass in solar masses
    mass2 : float
        Secondary mass in solar masses
    spin1z : float
         Dimensionless primary aligned spin component
    spin2z : float
         Dimensionless secondary aligned spin component
    snr : float
        Signal to noise ratio

    Returns
    -------
    str
        JSON formatted string storing ``HasNS`` and ``HasRemnant``
        probabilities

    Examples
    --------
    >>> em_bright.classifier_other(2.0, 1.0, 0.0, 0.0, 10.)
    '{"HasNS": 1.0, "HasRemnant": 1.0}'
    """
    p_ns, p_em = _em_bright(mass1, mass2, spin1z, spin2z)

    data = json.dumps({
        'HasNS': p_ns,
        'HasRemnant': p_em
    })
    return data


@app.task(shared=False)
def classifier_gstlal(mass1, mass2, spin1z, spin2z, snr):
    """Returns the probability of having a NS component and the probability of
    having non-zero disk mass in the detected event. This method will be using
    the data products obtained from the weekly supervised learning runs for
    injections campaigns. The data products are in pickle formatted
    RandomForestClassifier objects. The method predict_proba of these objects
    provides us the probabilities of the coalesence being EM-Bright and
    existence of neutron star in the binary.

    Parameters
    ----------
    mass1 : float
        Primary mass in solar masses
    mass2 : float
        Secondary mass in solar masses
    spin1z : float
         Dimensionless primary aligned spin component
    spin2z : float
         Dimensionless secondary aligned spin component
    snr : float
        Signal to noise ratio

    Returns
    -------
    str
        JSON formatted string storing ``HasNS`` and ``HasRemnant``
        probabilities

    Examples
    --------
    >>> em_bright.classifier_gstlal(2.0, 1.0, 0.0, 0.0, 10.)
    '{"HasNS": 1.0, "HasRemnant": 1.0}'
    """
    p_ns, p_em = em_bright.source_classification(
        mass1, mass2, spin1z, spin2z, snr,
        ns_classifier=NS_CLASSIFIER,
        emb_classifier=EM_CLASSIFIER)

    data = json.dumps({
        'HasNS': p_ns,
        'HasRemnant': p_em
    })
    return data



from unittest.mock import patch

import pkg_resources
import pytest

from ..util import resource_json
from .test_tasks_skymaps import toy_fits_filecontents  # noqa: F401
from .test_tasks_skymaps import toy_3d_fits_filecontents  # noqa: F401
from ..tasks import external_skymaps


true_heasarc_link = ('http://heasarc.gsfc.nasa.gov/FTP/fermi/data/gbm/'
                     + 'triggers/2017/bn170817529/current/')
true_skymap_link = true_heasarc_link + 'glg_healpix_all_bn170817529_v00.fit'


def mock_get_event(exttrig):
    return {'search': 'GRB'}


def mock_get_superevent(graceid):
    return resource_json(__name__, 'data/mock_superevent_object.json')


def mock_get_log(graceid):
    if graceid == 'S12345':
        return resource_json(__name__, 'data/gracedb_setrigger_log.json')
    elif graceid == 'E12345':
        return resource_json(__name__, 'data/gracedb_externaltrigger_log.json')
    else:
        raise ValueError


@pytest.fixture  # noqa: F811
def mock_download(monkeypatch, toy_3d_fits_filecontents):  # noqa: F811

    def download(filename, graceid):
        """Mocks GraceDB download functionality"""
        if graceid == 'S12345' and filename == 'bayestar.fits.gz':
            return toy_3d_fits_filecontents
        elif (graceid == 'E12345' and
              filename == ('nasa.gsfc.gcn_Fermi%23GBM_Gnd_Pos_2017-08-17'
                           + 'T12%3A41%3A06.47_524666471_57-431.xml')):
            return pkg_resources.resource_string(
                       __name__, 'data/externaltrigger_original_data.xml'
                   )
        else:
            raise ValueError

    monkeypatch.setattr('gwcelery.tasks.gracedb.download.run', download)


def mock_get_file_contents(monkeypatch, toy_fits_filecontents):  # noqa: F811
    """Mocks astropy get_file_contents functionality"""
    def get_file_contents(heasarc_link):
        assert heasarc_link == true_heasarc_link
        return toy_fits_filecontents

    monkeypatch.setattr(
        'astropy.utils.data.get_file_contents', get_file_contents)


@patch('gwcelery.tasks.skymaps.plot_allsky.run')
@patch('gwcelery.tasks.gracedb.upload.run')
@patch('gwcelery.tasks.external_skymaps.combine_skymaps.run')
@patch('gwcelery.tasks.gracedb.download')
@patch('gwcelery.tasks.external_skymaps.get_skymap_filename',
       return_value='fermi_skymap.fits.gz,0')
def test_create_combined_skymap(mock_get_skymap_filename,
                                mock_download,
                                mock_combine_skymaps, mock_upload,
                                mock_plot_allsky):
    """Test creating combined LVC and Fermi skymap"""
    # Run function under test
    external_skymaps.create_combined_skymap('S12345', 'E12345')
    mock_combine_skymaps.assert_called_once()
    mock_upload.assert_called()


@patch('gwcelery.tasks.gracedb.get_log', mock_get_log)
def test_get_skymap_filename():
    """Test getting the LVC skymap fits filename"""
    external_skymaps.get_skymap_filename('S12345')


@patch('gwcelery.tasks.gracedb.get_event', mock_get_event)
@patch('gwcelery.tasks.gracedb.get_superevent',
       return_value={'em_events': ['E12345']})
def test_external_trigger(mock_get_superevent, mock_download):
    """Test getting related em event for superevent"""
    assert external_skymaps.external_trigger('S12345') == 'E12345'


@patch('gwcelery.tasks.gracedb.get_log', mock_get_log)
def test_external_trigger_heasarc(mock_download):
    """Test retrieving HEASARC fits file link from GCN"""
    heasarc_link = external_skymaps.external_trigger_heasarc('E12345')
    assert heasarc_link == true_heasarc_link


@patch('urllib.request.urlopen')
def test_get_external_skymap(mock_urlopen):
    """Assert that the correct call to astropy.get_file_contents is used"""
    external_skymaps.get_external_skymap(true_heasarc_link, 'GRB')

    mock_urlopen.assert_called_once()


@patch('gwcelery.tasks.gracedb.upload.run')
@patch('gwcelery.tasks.skymaps.plot_allsky.run')
@patch('gwcelery.tasks.external_skymaps.get_external_skymap.run')
@patch('gwcelery.tasks.external_skymaps.external_trigger_heasarc.run')
def test_get_upload_external_skymap(mock_external_trigger_heasarc,
                                    mock_get_external_skymap,
                                    mock_plot_allsky,
                                    mock_upload):
    """Test that an external sky map is grabbed and uploaded."""
    graceid = 'E12345'
    external_skymaps.get_upload_external_skymap(graceid, 'GRB')
    mock_external_trigger_heasarc.assert_called_once()
    mock_get_external_skymap.assert_called_once()
    mock_upload.assert_called()


@patch('gwcelery.tasks.gracedb.upload.run')
@patch('gwcelery.tasks.skymaps.plot_allsky.run')
@patch('gwcelery.tasks.external_skymaps.get_external_skymap.run')
def test_get_upload_external_skymap_subgrb(mock_get_external_skymap,
                                           mock_plot_allsky,
                                           mock_upload):
    """Test that an external sky map is grabbed and uploaded."""
    graceid = 'E12345'
    external_skymaps.get_upload_external_skymap(
        graceid, 'SubGRB',
        ('https://gcn.gsfc.nasa.gov/notices_gbm_sub/' +
         'gbm_subthresh_604671025.728000_healpix.fits'))
    mock_get_external_skymap.assert_called_once()
    mock_upload.assert_called()


@pytest.mark.parametrize('ra,dec,error,pix',
                         [[0, 90, 0, 0],
                          [270, -90, .01, -1]])
def test_create_swift_skymap(ra, dec, error, pix):
    """Test created single pixel sky maps for Swift localization."""
    skymap = external_skymaps.create_external_skymap(ra, dec, error, 'Swift')
    assert skymap[pix] == 1


def test_create_fermi_skymap():
    """Test created single pixel sky maps for Swift localization."""
    ra, dec, error = 0, 90, 10
    external_skymaps.create_external_skymap(ra, dec, error, 'Fermi')


@patch('gwcelery.tasks.gracedb.upload.run')
@patch('gwcelery.tasks.skymaps.plot_allsky.run')
def test_create_upload_swift_skymap(mock_plot_allsky,
                                    mock_upload):
    """Test the creation and upload of sky maps for Swift localization."""
    event = {'graceid': 'E1234',
             'pipeline': 'Swift',
             'gpstime': 1259790538.77,
             'extra_attributes': {
                 'GRB': {
                     'trigger_id': 1234567,
                     'ra': 0,
                     'dec': 0,
                     'error_radius': 0}},
             'links': {
                 'self': 'https://gracedb.ligo.org/api/events/E356793'}}
    external_skymaps.create_upload_external_skymap(event, '111',
                                                   '2020-01-09T01:47:09')
    mock_upload.assert_called()
    mock_plot_allsky.assert_called_once()



"""Communication with GraceDB using the legacy GraceDB client."""
import functools
from http.client import HTTPException
from socket import gaierror

from ligo.gracedb import rest

from ..import app
from ..util import PromiseProxy

client = PromiseProxy(rest.GraceDb,
                      ('https://' + app.conf.gracedb_host + '/api/',),
                      {'fail_if_noauth': True, 'reload_certificate': True})


class RetryableHTTPError(rest.HTTPError):
    """Exception class for server-side HTTP errors that we should retry."""


def catch_retryable_http_errors(f):
    """Decorator to capture server-side errors that we should retry.

    We retry HTTP status 502 (Bad Gateway), 503 (Service Unavailable), and
    504 (Gateway Timeout).
    """
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except rest.HTTPError as e:
            if e.status in {429, 502, 503, 504}:
                raise RetryableHTTPError(e.status, e.reason, e.message)
            else:
                raise

    return wrapper


def task(*args, **kwargs):
    return app.task(*args, **kwargs,
                    autoretry_for=(gaierror, RetryableHTTPError,
                                   TimeoutError, HTTPException),
                    default_retry_delay=20.0, retry_backoff=True,
                    retry_kwargs=dict(max_retries=10))



from unittest.mock import call, patch

import pytest

from .test_tasks_skymaps import toy_fits_filecontents  # noqa: F401
from ..tasks import legacy_gracedb as gracedb
from ..tasks import raven


@pytest.mark.parametrize(
    'group,gracedb_id,pipelines,ext_search,tl,th',
    [['CBC', 'S1', ['Fermi', 'Swift'], 'GRB', -1, 5],
     ['Burst', 'S2', ['Fermi', 'Swift'], 'GRB', -60, 600],
     ['Burst', 'S3', ['SNEWS'], 'Supernova', -10, 10],
     ['CBC', 'E1', ['Fermi'], 'GRB', -5, 1]])
@patch('gwcelery.tasks.gracedb.create_label.run')
@patch('gwcelery.tasks.raven.raven_pipeline.run')
@patch('gwcelery.tasks.raven.search.run', return_value=[{'superevent_id': 'S5',
                                                         'graceid': 'E2'}])
@patch('gwcelery.tasks.raven.calculate_coincidence_far.run')
def test_coincidence_search(mock_calculate_coincidence_far,
                            mock_search, mock_raven_pipeline,
                            mock_create_label,
                            group, gracedb_id, pipelines, ext_search, tl, th):
    """Test that correct time windows are used for each RAVEN search."""
    alert_object = {'superevent_id': gracedb_id}
    if 'E' in gracedb_id:
        alert_object['group'] = 'External'
    raven.coincidence_search(gracedb_id, alert_object, group,
                             pipelines, ext_search)

    mock_search.assert_called_once_with(
        gracedb_id, alert_object, tl, th, group, pipelines, ext_search)
    mock_raven_pipeline.assert_called_once_with(
        [{'superevent_id': 'S5', 'graceid': 'E2'}],
        gracedb_id, alert_object, tl, th, group)


@pytest.mark.parametrize(
    'event_type,event_id', [['SE', 'S1234'], ['ExtTrig', 'E1234']])
@patch('ligo.raven.gracedb_events.ExtTrig')
@patch('ligo.raven.gracedb_events.SE')
@patch('ligo.raven.search.search')
def test_raven_search(mock_raven_search, mock_se_cls, mock_exttrig_cls,
                      event_type, event_id):
    """Test that correct input parameters are used for raven."""
    alert_object = {}
    if event_type == 'SE':
        alert_object['superevent_id'] = event_id

    # call raven search
    raven.search(event_id, alert_object)
    if event_id == 'S1234':
        mock_raven_search.assert_called_once_with(
            mock_se_cls(event_id, gracedb=gracedb.client), -5, 5,
            gracedb=gracedb.client, group=None, pipelines=[], searches=[])
    elif event_id == 'E1234':
        mock_raven_search.assert_called_once_with(
            mock_exttrig_cls(event_id, gracedb=gracedb.client), -5, 5,
            gracedb=gracedb.client, group=None, pipelines=[], searches=[])
    else:
        raise ValueError


@pytest.mark.parametrize('group', ['CBC', 'Burst'])
@patch('ligo.raven.search.calc_signif_gracedb')
def test_calculate_coincidence_far(
        mock_calc_signif, group):
    se = {'superevent_id': 'S1234'}
    ext = {'graceid': 'E4321',
           'pipeline': 'Fermi',
           'search': 'GRB',
           'labels': [],
           'far': None}
    if group == 'CBC':
        tl, th = -5, 1
    else:
        tl, th = -600, 60
    raven.calculate_coincidence_far(se, ext, tl, th)
    mock_calc_signif.assert_called_once_with(
        'S1234', 'E4321', tl, th,
        incl_sky=False, grb_search='GRB',
        gracedb=gracedb.client, far_grb=None)


@patch('ligo.raven.search.calc_signif_gracedb')
def test_calculate_coincidence_far_subgrb(mock_calc_signif):
    se = {'superevent_id': 'S1234'}
    ext = {'graceid': 'E4321',
           'pipeline': 'Fermi',
           'search': 'GRB',
           'labels': [],
           'far': 1e5}
    tl, th = -1, 10
    raven.calculate_coincidence_far(se, ext, tl, th)
    mock_calc_signif.assert_called_once_with(
        'S1234', 'E4321', tl, th,
        incl_sky=False, grb_search='GRB',
        gracedb=gracedb.client, far_grb=1e5)


@pytest.mark.parametrize('group', ['CBC', 'Burst'])  # noqa: F811
@patch('gwcelery.tasks.external_skymaps.get_skymap_filename',
       return_value='fermi_skymap.fits.gz')
@patch('ligo.raven.search.calc_signif_gracedb')
def test_calculate_spacetime_coincidence_far(
        mock_calc_signif, mock_get_skymap_filename, group):
    se = {'superevent_id': 'S1234'}
    ext = {'graceid': 'E4321',
           'pipeline': 'Fermi',
           'search': 'GRB',
           'labels': ['EXT_SKYMAP_READY', 'SKYMAP_READY'],
           'far': None}
    if group == 'CBC':
        tl, th = -5, 1
    else:
        tl, th = -600, 60
    raven.calculate_coincidence_far(se, ext, tl, th)
    mock_calc_signif.assert_called_once_with(
        'S1234', 'E4321', tl, th,
        incl_sky=True, grb_search='GRB',
        se_fitsfile='fermi_skymap.fits.gz',
        ext_fitsfile='fermi_skymap.fits.gz',
        gracedb=gracedb.client, far_grb=None)


def mock_get_labels(superevent_id):
    if superevent_id == 'S14':
        return {'ADVREQ'}
    else:
        return {}


def mock_coinc_far(*args):
    return {'temporal_coincidence_far': 1e-7,
            'spatiotemporal_coinc_far': None}


@pytest.mark.parametrize(
    'raven_search_results,graceid,tl,th,group',
    [[[{'graceid': 'E1', 'pipeline': 'GRB'}], 'S1', -5, 1, 'CBC'],
     [[{'superevent_id': 'S10', 'far': 1, 'preferred_event': 'G1'}],
        'E2', -1, 5, 'CBC'],
     [[{'graceid': 'E3', 'pipeline': 'GRB'},
       {'graceid': 'E4', 'pipeline': 'GRB'}], 'S2', -600, 60, 'Burst'],
     [[{'superevent_id': 'S11', 'far': 1, 'preferred_event': 'G2'},
       {'superevent_id': 'S12', 'far': .001, 'preferred_event': 'G3'}],
        'E5', -1, 5, 'CBC'],
     [[], 'S13', -1, 5, 'CBC'],
     [[{'graceid': 'E4', 'pipeline': 'GRB'}], 'S14', -1, 5, 'CBC']])
@patch('gwcelery.tasks.raven.trigger_raven_alert.run')
@patch('gwcelery.tasks.gracedb.get_labels', mock_get_labels)
@patch('gwcelery.tasks.raven.calculate_coincidence_far.run',
       return_value=mock_coinc_far())
@patch('gwcelery.tasks.gracedb.create_label.run')
def test_raven_pipeline(mock_create_label,
                        mock_calculate_coincidence_far,
                        mock_trigger_raven_alert,
                        raven_search_results, graceid, tl, th, group):
    """This function tests that the RAVEN pipeline runs correctly for scenarios
    where RAVEN finds nothing, a coincidence is found but does not pass
    threshold, when a coincidence is found but does pass threshold, and when
    multiple events are found.
    """
    alert_object = {'preferred_event': 'G1', 'pipeline': 'GRB', 'labels': []}
    for result in raven_search_results:
        result['labels'] = []
    if 'E' in graceid:
        alert_object['group'] = 'External'
    raven.raven_pipeline(raven_search_results, graceid, alert_object, tl, th,
                         group)

    coinc_calls = []
    label_calls = []
    if not raven_search_results:
        mock_calculate_coincidence_far.assert_not_called()
        mock_create_label.assert_not_called()
        return
    elif graceid.startswith('E'):
        result = raven.preferred_superevent(raven_search_results)[0]
        label_calls.append(call('EM_COINC', result['superevent_id']))
        coinc_calls.append(call(result, alert_object, tl, th))
        label_calls.append(call('EM_COINC', graceid))
    else:
        for result in raven_search_results:
            label_calls.append(call('EM_COINC', result['graceid']))
            coinc_calls.append(call(alert_object, result, tl, th))
            label_calls.append(call('EM_COINC', graceid))

    alert_calls = []
    if graceid.startswith('E'):
        alert_calls.append(call(
            mock_coinc_far(),
            result, graceid, alert_object, group))
    else:
        for result in raven_search_results:
            alert_calls.append(call(mock_coinc_far(),
                                    alert_object, graceid, result, group))
    mock_trigger_raven_alert.assert_has_calls(alert_calls, any_order=True)

    mock_calculate_coincidence_far.assert_has_calls(coinc_calls,
                                                    any_order=True)
    mock_create_label.assert_has_calls(label_calls, any_order=True)


@pytest.mark.parametrize(
    'raven_search_results, testnum',
    [[[{'superevent_id': 'S10', 'far': 1, 'preferred_event': 'G1'}], 1],
     [[{'superevent_id': 'S11', 'far': 1, 'preferred_event': 'G2'},
       {'superevent_id': 'S12', 'far': .001, 'preferred_event': 'G3'}], 2],
     [[{'superevent_id': 'S13', 'far': 1, 'preferred_event': 'G4'},
       {'superevent_id': 'S14', 'far': .0001, 'preferred_event': 'G5'},
       {'superevent_id': 'S15', 'far': .001, 'preferred_event': 'G6'}], 3]])
def test_preferred_superevent(raven_search_results, testnum):

    preferred_superevent = raven.preferred_superevent(raven_search_results)
    if testnum == 1:
        assert preferred_superevent == [{'superevent_id': 'S10', 'far': 1,
                                         'preferred_event': 'G1'}]
    if testnum == 2:
        assert preferred_superevent == [{'superevent_id': 'S12', 'far': .001,
                                         'preferred_event': 'G3'}]
    if testnum == 3:
        assert preferred_superevent == [{'superevent_id': 'S14', 'far': .0001,
                                         'preferred_event': 'G5'}]


def _mock_get_event(graceid):
    if graceid == "S1234":
        return {"superevent_id": "S1234",
                "preferred_event": "G000001",
                "far": 1e-5}
    elif graceid == "S2345":
        return {"superevent_id": "S2345",
                "preferred_event": "G000002",
                "far": 1e-10}
    elif graceid == "S2468":
        return {"superevent_id": "S2468",
                "preferred_event": "G000002",
                "far": 1e-4}
    elif graceid == "S5678":
        return {"superevent_id": "S5678",
                "preferred_event": "G000003",
                "far": 1e-5}
    elif graceid == "S8642":
        return {"superevent_id": "S8642",
                "preferred_event": "G000003",
                "far": 1e-3}
    elif graceid == "S9876":
        return {"superevent_id": "S9876",
                "preferred_event": "G000003",
                "far": 1e-6}
    elif graceid == 'E1':
        return {"graceid": "E1",
                "pipeline": 'Swift',
                "search": 'GRB',
                "labels": []}
    elif graceid == 'E2':
        return {"graceid": "E2",
                "pipeline": 'Fermi',
                "search": 'SubGRB',
                "labels": []}
    elif graceid == 'E3':
        return {"graceid": "E3",
                "pipeline": 'SNEWS',
                "search": 'Supernova',
                "labels": []}
    elif graceid == 'E4':
        return {"graceid": "E4",
                "pipeline": 'Fermi',
                "search": 'GRB',
                "labels": ['NOT_GRB']}
    else:
        raise AssertionError


def _mock_get_coinc_far(graceid):
    if graceid == "S1234":
        return {"temporal_coinc_far": 1e-12,
                "spatiotemporal_coinc_far": 1e-05}
    elif graceid == "S2468":
        return {"temporal_coinc_far": 1e-09,
                "spatiotemporal_coinc_far": None}
    elif graceid == "S5678":
        return {"temporal_coinc_far": 1e-15,
                "spatiotemporal_coinc_far": None}
    elif graceid == "S8642":
        return {"temporal_coinc_far": 1e-03,
                "spatiotemporal_coinc_far": None}
    elif graceid == "S9876":
        return {"temporal_coinc_far": 1e-07,
                "spatiotemporal_coinc_far": 1e-13}
    else:
        return {}


@pytest.mark.parametrize(
    'graceid,result_id,group,expected_result',
    [['S1234', 'E1', 'Burst', False],
     ['S2468', 'E1', 'CBC', False],
     ['S2468', 'E4', 'CBC', False],
     ['S2468', 'E2', 'CBC', False],
     ['S2345', 'E3', 'Burst', True],
     ['E1', 'S9876', 'CBC', True],
     ['E1', 'S2468', 'CBC', False],
     ['E2', 'S5678', 'CBC', False],
     ['E3', 'S8642', 'Burst', False],
     ['E3', 'S9876', 'Burst', True],
     ['E4', 'S5678', 'CBC', False]])
@patch('gwcelery.tasks.gracedb.get_labels', return_value={})
@patch('gwcelery.tasks.gracedb.update_superevent')
@patch('gwcelery.tasks.gracedb.create_label.run')
def test_trigger_raven_alert(mock_create_label, mock_update_superevent,
                             mock_get_labels,
                             graceid, result_id, group, expected_result):
    if graceid.startswith('E'):
        superevent_id = result_id
        ext_id = graceid
    else:
        superevent_id = graceid
        ext_id = result_id
    superevent = _mock_get_event(superevent_id)
    coinc_far_json = _mock_get_coinc_far(superevent_id)
    ext_event = _mock_get_event(ext_id)
    preferred_id = superevent['preferred_event']
    raven.trigger_raven_alert(coinc_far_json, superevent,
                              graceid, ext_event, group)

    if expected_result:
        label_calls = [call('RAVEN_ALERT', superevent_id),
                       call('RAVEN_ALERT', ext_id),
                       call('RAVEN_ALERT', preferred_id)]
        if ext_event['pipeline'] == 'SNEWS':
            time_coinc_far = space_coinc_far = None
        else:
            time_coinc_far = coinc_far_json['temporal_coinc_far']
            space_coinc_far = coinc_far_json['spatiotemporal_coinc_far']
        mock_update_superevent.assert_called_once_with(
            superevent_id,
            em_type=ext_id,
            time_coinc_far=time_coinc_far,
            space_coinc_far=space_coinc_far)
        mock_create_label.assert_has_calls(label_calls)
    else:
        mock_create_label.assert_not_called()



"""Create and upload external sky maps."""
from astropy import units as u
from astropy.coordinates import ICRS, SkyCoord
from astropy_healpix import HEALPix, pixel_resolution_to_nside
from celery import group
#  import astropy.utils.data
import numpy as np
from ligo.skymap.io import fits
from ligo.skymap.tool import ligo_skymap_combine
import gcn
import healpy as hp
import lxml.etree
import re
import ssl
import urllib

from ..import app
from . import gracedb
from . import skymaps
from ..util.cmdline import handling_system_exit
from ..util.tempfile import NamedTemporaryFile
from ..import _version


def create_combined_skymap(se_id, ext_id):
    """Creates and uploads the combined LVC-Fermi skymap.

    This also uploads the external trigger skymap to the external trigger
    GraceDB page.
    """
    se_skymap_filename = get_skymap_filename(se_id)
    ext_skymap_filename = get_skymap_filename(ext_id)
    new_skymap_filename = re.findall(r'(.*).fits.gz', se_skymap_filename)[0]

    #  FIXME: put download functions in canvas
    se_skymap = gracedb.download(se_skymap_filename, se_id)
    ext_skymap = gracedb.download(ext_skymap_filename, ext_id)
    message = 'Combined LVC-external sky map using {0} and {1}'.format(
        se_skymap_filename, ext_skymap_filename)
    message_png = (
        'Mollweide projection of <a href="/api/events/{graceid}/files/'
        '{filename}">{filename}</a>').format(
            graceid=se_id, filename=new_skymap_filename + '-ext.fits.gz')

    (
        combine_skymaps.si(se_skymap, ext_skymap)
        |
        group(
            gracedb.upload.s(new_skymap_filename + '-ext.fits.gz', se_id,
                             message, ['sky_loc', 'public']),

            skymaps.plot_allsky.s()
            |
            gracedb.upload.s(new_skymap_filename + '-ext.png', se_id,
                             message_png, ['sky_loc', 'ext_coinc', 'public'])
        )
    ).delay()


@app.task(autoretry_for=(ValueError,), retry_backoff=10,
          retry_backoff_max=600)
def get_skymap_filename(graceid):
    """Get the skymap fits filename.

    If not available, will try again 10 seconds later, then 20, then 40, etc.
    until up to 10 minutes after initial attempt.
    """
    gracedb_log = gracedb.get_log(graceid)
    if 'S' in graceid:
        for message in reversed(gracedb_log):
            filename = message['filename']
            if filename.endswith('.fits.gz'):
                return filename
    else:
        for message in reversed(gracedb_log):
            filename = message['filename']
            if (filename.endswith('.fits') or filename.endswith('.fit') or
                    filename.endswith('.fits.gz')):
                return filename
    raise ValueError('No skymap available for {0} yet.'.format(graceid))


@app.task(shared=False)
def combine_skymaps(skymap1filebytes, skymap2filebytes):
    """This task combines the two input skymaps, in this case the external
    trigger skymap and the LVC skymap and writes to a temporary output file. It
    then returns the contents of the file as a byte array.
    """
    with NamedTemporaryFile(mode='rb', suffix='.fits.gz') as combinedskymap, \
            NamedTemporaryFile(content=skymap1filebytes) as skymap1file, \
            NamedTemporaryFile(content=skymap2filebytes) as skymap2file, \
            handling_system_exit():
        ligo_skymap_combine.main([skymap1file.name,
                                  skymap2file.name, combinedskymap.name])
        return combinedskymap.read()


@app.task(shared=False)
def external_trigger(graceid):
    """Returns the associated external trigger GraceDB ID."""
    em_events = gracedb.get_superevent(graceid)['em_events']
    if len(em_events):
        for exttrig in em_events:
            if gracedb.get_event(exttrig)['search'] == 'GRB':
                return exttrig
    raise ValueError('No associated GRB EM event(s) for {0}.'.format(graceid))


@app.task(shared=False)
def external_trigger_heasarc(external_id):
    """Returns the HEASARC fits file link."""
    gracedb_log = gracedb.get_log(external_id)
    for message in gracedb_log:
        if 'Original Data' in message['comment']:
            filename = message['filename']
            xmlfile = gracedb.download(urllib.parse.quote(filename),
                                       external_id)
            root = lxml.etree.fromstring(xmlfile)
            heasarc_url = root.find('./What/Param[@name="LightCurve_URL"]'
                                    ).attrib['value']
            return re.sub(r'quicklook(.*)', 'current/', heasarc_url)
    raise ValueError('Not able to retrieve HEASARC link for {0}.'.format(
        external_id))


@app.task(autoretry_for=(urllib.error.HTTPError,), retry_backoff=10,
          retry_backoff_max=600)
def get_external_skymap(link, search):
    """Download the Fermi sky map fits file and return the contents as a byte
    array. If GRB, will construct a HEASARC url, while if SubGRB, will use the
    link directly.

    If not available, will try again 10 seconds later, then 20, then 40, etc.
    until up to 10 minutes after initial attempt.
    """
    if search == 'GRB':
        # if Fermi GRB, determine final HEASARC link
        trigger_id = re.sub(r'.*\/(\D+?)(\d+)(\D+)\/.*', r'\2', link)
        skymap_name = 'glg_healpix_all_bn{0}_v00.fit'.format(trigger_id)
        skymap_link = link + skymap_name
    elif search == 'SubGRB':
        skymap_link = link
    #  FIXME: Under Anaconda on the LIGO Caltech computing cluster, Python
    #  (and curl, for that matter) fail to negotiate TLSv1.2 with
    #  heasarc.gsfc.nasa.gov
    context = ssl.create_default_context()
    context.options |= ssl.OP_NO_TLSv1_3
    #  return astropy.utils.data.get_file_contents(
    #      (skymap_link), encoding='binary', cache=False)
    return urllib.request.urlopen(skymap_link, context=context).read()


@app.task(autoretry_for=(urllib.error.HTTPError, urllib.error.URLError,
          ValueError,), retry_backoff=10, retry_backoff_max=1200)
def get_upload_external_skymap(graceid, search, skymap_link=None):
    """If a Fermi sky map is not uploaded yet, tries to download one and upload
    to external event. If sky map is not available, passes so that this can be
    re-run the next time an update GCN notice is received. If GRB, will
    construct a HEASARC url, while if SubGRB, will use the link directly.
    """
    try:
        filename = get_skymap_filename(graceid)
        if 'glg_healpix_all_bn_v00.fit' in filename:
            return
    except ValueError:
        pass

    if search == 'GRB':
        external_skymap_canvas = (
            external_trigger_heasarc.si(graceid)
            |
            get_external_skymap.s(search)
        )
    elif search == 'SubGRB':
        external_skymap_canvas = get_external_skymap.si(skymap_link, search)

    skymap_filename = 'glg_healpix_all_bn_v00'

    message = (
        'Mollweide projection of <a href="/api/events/{graceid}/files/'
        '{filename}">{filename}</a>').format(
            graceid=graceid, filename=skymap_filename + '.fits')

    (
        external_skymap_canvas
        |
        group(
            gracedb.upload.s(
                skymap_filename + '.fits',
                graceid,
                'Official sky map from Fermi analysis.',
                ['sky_loc']),

            skymaps.plot_allsky.s()
            |
            gracedb.upload.s(skymap_filename + '.png',
                             graceid,
                             message,
                             ['sky_loc'])
        )
        |
        gracedb.create_label.si('EXT_SKYMAP_READY', graceid)
    ).delay()


def create_external_skymap(ra, dec, error, pipeline, notice_type=111):
    """Create a sky map, either a gaussian or a single
    pixel sky map, given an RA, dec, and error radius.

    If from Fermi, convolves the sky map with both a core and
    tail Gaussian and then sums these to account for systematic
    effects as measured in :doi:`10.1088/0067-0049/216/2/32`

    If from Swift, converts the error radius from that containing 90% of the
    credible region to ~68% (see description of Swift error
    here:`https://gcn.gsfc.nasa.gov/swift.html#tc7`)

    Parameters
    ----------
    ra : float
        right ascension in deg
    dec: float
        declination in deg
    error: float
        error radius in deg

    Returns
    -------
    skymap : numpy array
        sky map array

    """
    max_nside = 2048
    if error:
        # Correct 90% containment to 1-sigma for Swift
        if pipeline == 'Swift':
            error /= np.sqrt(-2 * np.log1p(-.9))
        error_radius = error * u.deg
        nside = pixel_resolution_to_nside(error_radius, round='up')
    else:
        nside = np.inf
    if nside >= max_nside:
        nside = max_nside

        #  Find the one pixel the event can localized to
        hpx = HEALPix(nside, 'ring', frame=ICRS())
        skymap = np.zeros(hpx.npix)
        ind = hpx.lonlat_to_healpix(ra * u.deg, dec * u.deg)
        skymap[ind] = 1.
    else:
        #  If larger error, create gaussian sky map
        hpx = HEALPix(nside, 'ring', frame=ICRS())
        ipix = np.arange(hpx.npix)

        #  Evaluate Gaussian.
        center = SkyCoord(ra * u.deg, dec * u.deg)
        distance = hpx.healpix_to_skycoord(ipix).separation(center)
        skymap = np.exp(-0.5 * np.square(distance / error_radius).to_value(
            u.dimensionless_unscaled))
        skymap /= skymap.sum()
    if pipeline == 'Fermi':
        # Correct for Fermi systematics based on recommendations from GBM team
        # Convolve with both a narrow core and wide tail Gaussian with error
        # radius determined by the scales respectively, each comprising a
        # fraction determined by the weights respectively
        if notice_type == gcn.NoticeType.FERMI_GBM_FLT_POS:
            # Flight notice
            # Values from first row of Table 7
            weights = [0.897, 0.103]
            scales = [7.52, 55.6]
        elif notice_type == gcn.NoticeType.FERMI_GBM_GND_POS:
            # Ground notice
            # Values from first row of Table 3
            weights = [0.804, 0.196]
            scales = [3.72, 13.7]
        elif notice_type == gcn.NoticeType.FERMI_GBM_FIN_POS:
            # Final notice
            # Values from second row of Table 3
            weights = [0.900, 0.100]
            scales = [3.71, 14.3]
        else:
            raise AssertionError(
                'Need to provide a supported Fermi notice type')
        skymap = sum(
            weight * hp.sphtfunc.smoothing(skymap, sigma=np.radians(scale))
            for weight, scale in zip(weights, scales))

    return skymap


def write_to_fits(skymap, event, notice_type, notice_date):
    """Write external sky map fits file, populating the
    header with relevant info.

    Parameters
    ----------
    skymap : numpy array
        sky map array
    event : dict
        Dictionary of Swift external event

    Returns
    -------
    skymap fits : bytes array
        bytes array of sky map

    """
    notice_type_dict = {
        '53': 'INTEGRAL_WAKEUP',
        '54': 'INTEGRAL_REFINED',
        '55': 'INTEGRAL_OFFLINE',
        '60': 'SWIFT_BAT_GRB_ALERT',
        '61': 'SWIFT_BAT_GRB_POSITION',
        '105': 'AGILE_MCAL_ALERT',
        '110': 'FERMI_GBM_ALERT',
        '111': 'FERMI_GBM_FLT_POS',
        '112': 'FERMI_GBM_GND_POS',
        '115': 'FERMI_GBM_FINAL_POS',
        '131': 'FERMI_GBM_SUBTHRESHOLD'}

    if notice_type is None:
        msgtype = event['pipeline'] + '_LVK_TARGETED_SEARCH'
    else:
        msgtype = notice_type_dict[str(notice_type)]

    gcn_id = event['extra_attributes']['GRB']['trigger_id']
    with NamedTemporaryFile(suffix='.fits.gz') as f:
        fits.write_sky_map(f.name, skymap,
                           objid=gcn_id,
                           url=event['links']['self'],
                           instruments=event['pipeline'],
                           gps_time=event['gpstime'],
                           msgtype=msgtype,
                           msgdate=notice_date,
                           creator='gwcelery',
                           origin='LIGO-VIRGO-KAGRA',
                           vcs_version=_version.get_versions()['version'],
                           history='file only for internal use')
        with open(f.name, 'rb') as file:
            return file.read()


@app.task(shared=False)
def create_upload_external_skymap(event, notice_type, notice_date):
    """Create and upload external sky map using
    RA, dec, and error radius information.

    Parameters
    ----------
    event : dict
        Dictionary of Swift external event

    """
    graceid = event['graceid']
    skymap_filename = event['pipeline'].lower() + '_skymap.fits.gz'

    ra = event['extra_attributes']['GRB']['ra']
    dec = event['extra_attributes']['GRB']['dec']
    error = event['extra_attributes']['GRB']['error_radius']
    pipeline = event['pipeline']
    skymap = create_external_skymap(ra, dec, error, pipeline, notice_type)

    skymap_data = write_to_fits(skymap, event, notice_type, notice_date)

    message = (
        'Mollweide projection of <a href="/api/events/{graceid}/files/'
        '{filename}">{filename}</a>').format(
            graceid=graceid, filename=skymap_filename)

    (
        gracedb.upload.si(skymap_data,
                          skymap_filename,
                          graceid,
                          'Sky map created from GCN RA, dec, and error.',
                          ['sky_loc'])
        |
        skymaps.plot_allsky.si(skymap_data, ra=ra, dec=dec)
        |
        gracedb.upload.s(event['pipeline'].lower() + '_skymap.png',
                         graceid,
                         message,
                         ['sky_loc'])
        |
        gracedb.create_label.si('EXT_SKYMAP_READY', graceid)
    ).delay()



"""Application configuration for ``gracedb-test.ligo.org``.

Inherits all settings from :mod:`gwcelery.conf.playground`, with the exceptions
below.
"""
from . import *  # noqa: F401, F403

lvalert_host = 'lvalert-test.cgca.uwm.edu'
"""LVAlert host."""

gracedb_host = 'gracedb-test.ligo.org'
"""GraceDB host."""

sentry_environment = 'test'
"""Record this `environment tag
<https://docs.sentry.io/enriching-error-data/environments/>`_ in Sentry log
messages."""

mock_events_simulate_multiple_uploads = True
"""If True, then upload each mock event several times in rapid succession with
random jitter in order to simulate multiple pipeline uploads."""



"""Tasks to send and receive Gamma-ray Coordinates Network [GCN]_ notices.

References
----------
.. [GCN] https://gcn.gsfc.nasa.gov

"""
import html
import difflib
import urllib.parse

from comet.utility.xml import xml_document
import gcn
from gcn import get_notice_type, NoticeType
import lxml.etree
from twisted.internet import reactor

from ..voevent.signals import voevent_received
from ..import app
from .core import DispatchHandler
from . import gracedb


class _VOEventDispatchHandler(DispatchHandler):

    def process_args(self, event):
        notice_type = get_notice_type(event.element)

        # Just cast to enum for prettier log messages
        try:
            notice_type = NoticeType(notice_type)
        except ValueError:
            pass

        return notice_type, (event.raw_bytes,), {}


handler = _VOEventDispatchHandler()
r"""Function decorator to register a handler callback for specified GCN notice
types. The decorated function is turned into a Celery task, which will be
automatically called whenever a matching GCN notice is received.

Parameters
----------
\*keys
    List of GCN notice types to accept
\*\*kwargs
    Additional keyword arguments for :meth:`celery.Celery.task`.

Examples
--------
Declare a new handler like this::

    @gcn.handler(gcn.NoticeType.FERMI_GBM_GND_POS,
                 gcn.NoticeType.FERMI_GBM_FIN_POS)
    def handle_fermi(payload):
        root = lxml.etree.fromstring(payload)
        # do work here...
"""


@voevent_received.connect
def _on_voevent_received(xml_document, **kwargs):
    handler.dispatch(xml_document)


class SendingError(RuntimeError):
    """A generic error associated with sending VOEvents."""


@app.task(autoretry_for=(SendingError,), bind=True, default_retry_delay=20.0,
          ignore_result=True, queue='voevent', retry_backoff=True,
          retry_kwargs=dict(max_retries=10), shared=False)
def send(self, message):
    """Send a VOEvent to GCN.

    This task will be retried several times if the VOEvent cannot be sent. See
    the Raises section below for circumstances that cause a retry.

    Parameters
    ----------
    message : bytes
        The raw VOEvent file contents.

    Raises
    ------
    SendingError
        If the VOEvent could not be sent because there were no network peers
        connected to the VOEvent broadcaster.

    """
    broadcasters = self.app.conf['voevent_broadcaster_factory'].broadcasters
    if broadcasters:
        event = xml_document(message)
        for broadcaster in broadcasters:
            reactor.callFromThread(broadcaster.send_event, event)
    elif self.app.conf['voevent_broadcaster_whitelist']:
        raise SendingError('Not sending the event because there are no '
                           'subscribers connected to the GCN broker.')


@handler(gcn.NoticeType.LVC_EARLY_WARNING,
         gcn.NoticeType.LVC_PRELIMINARY,
         gcn.NoticeType.LVC_INITIAL,
         gcn.NoticeType.LVC_UPDATE,
         gcn.NoticeType.LVC_RETRACTION,
         bind=True, shared=False)
def validate(self, payload):
    """Validate LIGO/Virgo GCN notices.

    Check that the contents of a public LIGO/Virgo GCN matches the original
    VOEvent in GraceDB.

    Notes
    -----
    If the VOEvent broadcaster is disabled by setting
    :obj:`~gwcelery.conf.voevent_broadcaster_whitelist` to an empty list, then
    this task becomes a no-op.

    """
    if not self.app.conf['voevent_broadcaster_whitelist']:
        return

    root = lxml.etree.fromstring(payload)

    # Which GraceDB ID does this refer to?
    graceid = root.find("./What/Param[@name='GraceID']").attrib['value']

    # Which VOEvent does this refer to?
    u = urllib.parse.urlparse(root.attrib['ivorn'])
    local_id = u.fragment
    filename = local_id + '.xml'

    # Download and parse original VOEvent
    orig = gracedb.download(filename, graceid)

    # Create a diff of the two VOEvents.
    diff = ''.join(
        difflib.unified_diff(
            *(
                [
                    line.decode('ascii', 'surrogateescape')
                    for line in contents.splitlines(keepends=True)
                ]
                for contents in (orig, payload)
            ),
            fromfile='{} (sent)'.format(filename),
            tofile='{} (received)'.format(filename)
        )
    )

    if diff:
        # Write a log message to indicate that the event differed.
        msg = 'VOEvent received from GCN differs from what we sent.'
        gracedb.upload.delay(
            None, None, graceid,
            '{}<pre>{}</pre>'.format(msg, html.escape(diff)),
            ['em_follow'])
        raise ValueError('{}\n\n{}'.format(msg, diff))
    else:
        # Tag the VOEvent to indicate that it was received correctly.
        gracedb.create_tag.delay(filename, 'gcn_received', graceid)



"""All Celery tasks are declared in submodules of this module."""


def _import_all_submodules(module=vars()):
    import importlib
    import pkgutil

    try:
        path = module['__path__']
    except KeyError:
        # not a package, does not have submodules
        return
    for _, name, _ in pkgutil.iter_modules(path):
        submodule = importlib.import_module('.' + name, module['__name__'])
        module[name] = submodule
        _import_all_submodules(vars(submodule))


_import_all_submodules()
